#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
import os
from mitosheet.step_performers.graph_steps.graph_utils import BAR
from mitosheet.transpiler.transpile_utils import NEWLINE_TAB, TAB, NEWLINE
import pytest
import pandas as pd

from mitosheet.api.get_parameterizable_params import get_parameterizable_params
from mitosheet.transpiler.transpile import transpile
from mitosheet.tests.test_utils import create_mito_wrapper_with_data, create_mito_wrapper
from mitosheet.tests.decorators import pandas_post_1_2_only, python_post_3_6_only
from mitosheet.types import FC_NUMBER_GREATER

def test_transpile_single_column():
    mito = create_mito_wrapper_with_data(['abc'])
    mito.set_formula('=A', 0, 'B', add_column=True)

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        '',
        "df1['B'] = df1[\'A\']", 
        ''
    ]


def test_transpile_multiple_columns_no_relationship():
    mito = create_mito_wrapper_with_data(['abc'])
    mito.add_column(0, 'B')
    mito.add_column(0, 'C')
    print(mito.transpiled_code)
    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        '',
        'df1[\'B\'] = 0', 
        '',
        'df1[\'C\'] = 0', 
        ''
    ]

def test_transpile_columns_in_each_sheet():
    mito = create_mito_wrapper_with_data(['abc'], sheet_two_A_data=['abc'])
    mito.add_column(0, 'B')
    mito.add_column(1, 'B')

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        '',
        'df1[\'B\'] = 0',
        '',
        'df2[\'B\'] = 0',
        ''
    ]

def test_transpile_multiple_columns_linear():
    mito = create_mito_wrapper_with_data(['abc'])
    mito.set_formula('=A', 0, 'B', add_column=True)
    mito.set_formula('=B', 0, 'C', add_column=True)

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        '',
        'df1[\'B\'] = df1[\'A\']',
        '',
        'df1[\'C\'] = df1[\'B\']',
        '',
    ]

COLUMN_HEADERS = [
    ('ABC'),
    ('ABC_D'),
    ('ABC_DEF'),
    ('ABC_123'),
    ('ABC_HAHA_123'),
    ('ABC_HAHA-123'),
    ('---data---'),
    ('---da____ta---'),
    ('--'),
]
@pytest.mark.parametrize("column_header", COLUMN_HEADERS)
def test_transpile_column_headers_non_alphabet(column_header):
    mito = create_mito_wrapper_with_data(['abc'])
    mito.set_formula('=A', 0, column_header, add_column=True)

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        '',
        f'df1[\'{column_header}\'] = df1[\'A\']', 
        '',
    ]


COLUMN_HEADERS = [
    ('ABC'),
    ('ABC_D'),
    ('ABC_DEF'),
    ('ABC_123'),
    ('ABC_HAHA_123'),
    ('ABC_HAHA-123'),
    ('---data---'),
    ('---da____ta---'),
    ('--'),
]
@pytest.mark.parametrize("column_header", COLUMN_HEADERS)
def test_transpile_column_headers_non_alphabet_multi_sheet(column_header):
    mito = create_mito_wrapper_with_data(['abc'], sheet_two_A_data=['abc'])
    mito.set_formula('=A', 0, column_header, add_column=True)
    mito.set_formula('=A', 1, column_header, add_column=True)

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        '',
        f'df1[\'{column_header}\'] = df1[\'A\']', 
        '',
        f'df2[\'{column_header}\'] = df2[\'A\']', 
        '',
    ]

def test_preserves_order_columns():
    mito = create_mito_wrapper_with_data(['abc'])
    # Topological sort will currently display this in C, B order
    mito.add_column(0, 'B')
    mito.add_column(0, 'C')

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        '',
        'df1[\'B\'] = 0',
        '',
        'df1[\'C\'] = 0',
        '',
    ]

def test_transpile_delete_columns():
    df1 = pd.DataFrame(data={'A': [1], 'B': [101], 'C': [11]})
    mito = create_mito_wrapper(df1)
    mito.delete_columns(0, ['C', 'B'])

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        '',
        'df1.drop([\'C\', \'B\'], axis=1, inplace=True)',
        '',
    ]


# TESTING OPTIMIZATION

def test_removes_unedited_formulas_for_unedited_sheets():
    df1 = pd.DataFrame(data={'A': [1], 'B': [101], 'C': [11]})
    df2 = pd.DataFrame(data={'A': [1], 'B': [101], 'C': [11]})
    mito = create_mito_wrapper(df1, df2)

    mito.set_formula('=C', 0, 'D', add_column=True)
    mito.set_formula('=C', 1, 'D', add_column=True)

    mito.merge_sheets('lookup', 0, 1, [['A', 'A']], ['A', 'B', 'C', 'D'], ['A', 'B', 'C', 'D'])

    mito.set_formula('=C + 1', 1, 'D')

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        '',
        "df1['D'] = df1[\'C\']", 
        '',
        "df2['D'] = df2[\'C\']", 
        '',
        'temp_df = df2.drop_duplicates(subset=[\'A\']) # Remove duplicates so lookup merge only returns first match', 
        'df_merge = df1.merge(temp_df, left_on=[\'A\'], right_on=[\'A\'], how=\'left\', suffixes=[\'_df1\', \'_df2\'])',
        '',
        'df2[\'D\'] = df2[\'C\'] + 1',
        '',
    ]


def test_mulitple_merges_no_formula_steps():
    df1 = pd.DataFrame(data={'A': [1], 'B': [101], 'C': [11]})
    df2 = pd.DataFrame(data={'A': [1], 'B': [101], 'C': [11]})
    mito = create_mito_wrapper(df1, df2)
    mito.merge_sheets('lookup', 0, 1, [['A', 'A']], ['A', 'B', 'C'], ['A', 'B', 'C'])
    mito.merge_sheets('lookup', 0, 1, [['A', 'A']], ['A', 'B', 'C'], ['A', 'B', 'C'])
    mito.merge_sheets('lookup', 0, 1, [['A', 'A']], ['A', 'B', 'C'], ['A', 'B', 'C'])


    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        '',
        'temp_df = df2.drop_duplicates(subset=[\'A\']) # Remove duplicates so lookup merge only returns first match', 
        'df_merge = df1.merge(temp_df, left_on=[\'A\'], right_on=[\'A\'], how=\'left\', suffixes=[\'_df1\', \'_df2\'])',
        '',
        'temp_df = df2.drop_duplicates(subset=[\'A\']) # Remove duplicates so lookup merge only returns first match', 
        'df_merge_1 = df1.merge(temp_df, left_on=[\'A\'], right_on=[\'A\'], how=\'left\', suffixes=[\'_df1\', \'_df2\'])',
        '',
        'temp_df = df2.drop_duplicates(subset=[\'A\']) # Remove duplicates so lookup merge only returns first match', 
        'df_merge_2 = df1.merge(temp_df, left_on=[\'A\'], right_on=[\'A\'], how=\'left\', suffixes=[\'_df1\', \'_df2\'])',
        '',
    ]

def test_optimization_with_other_edits():
    df1 = pd.DataFrame(data={'A': [1], 'B': [101], 'C': [11]})
    df2 = pd.DataFrame(data={'A': [1], 'B': [101], 'C': [11]})
    mito = create_mito_wrapper(df1, df2)
    mito.add_column(0, 'D')
    mito.set_formula('=A', 0, 'D')
    mito.merge_sheets('lookup', 0, 1, [['A', 'A']], ['A', 'B', 'C', 'D'], ['A', 'B', 'C'])
    mito.add_column(0, 'AAA')
    mito.delete_columns(0, ['AAA'])

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        '',
        "df1['D'] = df1[\'A\']", 
        '',
        'temp_df = df2.drop_duplicates(subset=[\'A\']) # Remove duplicates so lookup merge only returns first match', 
        'df_merge = df1.merge(temp_df, left_on=[\'A\'], right_on=[\'A\'], how=\'left\', suffixes=[\'_df1\', \'_df2\'])',
        '',
    ]


def test_transpile_does_no_initial():
    df1 = pd.DataFrame(data={'First Name': ['Nate', 'Nate'], 123: ['Rush', 'Jack'], True: ['1', '2']})
    mito = create_mito_wrapper(df1)

    assert len(mito.transpiled_code) == 0

    
def test_transpile_reorder_column():
    df1 = pd.DataFrame(data={'A': ['aaron'], 'B': ['jon']})
    mito = create_mito_wrapper(df1)
    mito.reorder_column(0, 'A', 1)

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        '',
        'df1_columns = [col for col in df1.columns if col != \'A\']',
        'df1_columns.insert(1, \'A\')',
        'df1 = df1[df1_columns]',
        '',
    ]

def test_transpile_two_column_reorders():
    df1 = pd.DataFrame(data={'A': ['aaron'], 'B': ['jon']})
    mito = create_mito_wrapper(df1)
    mito.reorder_column(0, 'A', 1)
    mito.reorder_column(0, 'B', 1)

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        '',
        'df1_columns = [col for col in df1.columns if col != \'A\']',
        'df1_columns.insert(1, \'A\')',
        'df1 = df1[df1_columns]',
        '',
        'df1_columns = [col for col in df1.columns if col != \'B\']',
        'df1_columns.insert(1, \'B\')',
        'df1 = df1[df1_columns]',
        '',
    ]

def test_transpile_reorder_column_invalid():
    df1 = pd.DataFrame(data={'A': ['aaron'], 'B': ['jon']})
    mito = create_mito_wrapper(df1)
    mito.reorder_column(0, 'A', 5)

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        '',
        'df1_columns = [col for col in df1.columns if col != \'A\']',
        'df1_columns.insert(1, \'A\')',
        'df1 = df1[df1_columns]',
        '',
    ]

def test_transpile_merge_then_sort():
    df1 = pd.DataFrame(data={'Name': ["Aaron", "Nate"], 'Number': [123, 1]})
    df2 = pd.DataFrame(data={'Name': ["Aaron", "Nate"], 'Sign': ['Gemini', "Tarus"]})
    mito = create_mito_wrapper(df1, df2)
    mito.merge_sheets('lookup', 0, 1, [['Name', 'Name']], list(df1.keys()), list(df2.keys()))
    mito.sort(2, 'Number', 'ascending')

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        '',
        'temp_df = df2.drop_duplicates(subset=[\'Name\']) # Remove duplicates so lookup merge only returns first match',
        'df_merge = df1.merge(temp_df, left_on=[\'Name\'], right_on=[\'Name\'], how=\'left\', suffixes=[\'_df1\', \'_df2\'])',
        '',
        'df_merge = df_merge.sort_values(by=\'Number\', ascending=True, na_position=\'first\')',
        '',
    ]

def test_transpile_multiple_pandas_imports_combined(tmp_path):
    tmp_file = str(tmp_path / 'txt.csv')
    df1 = pd.DataFrame(data={'Name': ["Aaron", "Nate"], 'Number': [123, 1]})
    df1.to_csv(tmp_file, index=False)
    mito = create_mito_wrapper(df1)
    mito.simple_import([tmp_file])
    mito.add_column(0, 'A', -1)
    mito.simple_import([tmp_file])
    mito.add_column(1, 'A', -1)
    mito.simple_import([tmp_file])

    assert len(mito.optimized_code_chunks) == 3
    assert 'import pandas as pd' in mito.transpiled_code
    assert len([c for c in mito.transpiled_code if c == 'import pandas as pd']) == 1

def test_transpile_as_function_no_params(tmp_path):
    tmp_file = str(tmp_path / 'txt.csv')
    df1 = pd.DataFrame({'A': [1], 'B': [2]})
    df1.to_csv(tmp_file, index=False)

    mito = create_mito_wrapper()
    mito.simple_import([tmp_file])
    mito.code_options_update({'as_function': True, 'import_custom_python_code': False, 'import_custom_python_code': False, 'call_function': True, 'function_name': 'function', 'function_params': {}})

    print(mito.transpiled_code)

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        "import pandas as pd",
        "",
        "def function():",
        f"{TAB}txt = pd.read_csv(r'{tmp_file}')",
        f'{TAB}',
        f"{TAB}return txt",
        "",
        "txt = function()"
    ]

def test_transpile_as_function_no_call(tmp_path):
    tmp_file = str(tmp_path / 'txt.csv')
    df1 = pd.DataFrame({'A': [1], 'B': [2]})
    df1.to_csv(tmp_file, index=False)

    mito = create_mito_wrapper()
    mito.simple_import([tmp_file])
    mito.code_options_update_no_check_transpiled({'as_function': True, 'import_custom_python_code': False, 'call_function': False, 'function_name': 'function', 'function_params': {}})


    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        "import pandas as pd",
        "",
        "def function():",
        f"{TAB}txt = pd.read_csv(r'{tmp_file}')",
        f'{TAB}',
        f"{TAB}return txt",
        ""
    ]

def test_transpile_as_function_df_params():
    mito = create_mito_wrapper(pd.DataFrame({'A': [1]}), arg_names=['df1'])
    mito.add_column(0, 'B')
    mito.code_options_update({'as_function': True, 'import_custom_python_code': False, 'call_function': True, 'function_name': 'function', 'function_params': {}})

    print(mito.transpiled_code)
    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        '',
        'def function(df1):',
        f"{TAB}df1['B'] = 0",
        f'{TAB}',
        f"{TAB}return df1",
        "",
        "df1 = function(df1)"
    ]

def test_transpile_as_function_string_params():
    tmp_file = 'txt.csv'
    df1 = pd.DataFrame({'A': [1], 'B': [2]})
    df1.to_csv(tmp_file, index=False)

    mito = create_mito_wrapper(str(tmp_file), arg_names=[f"'{tmp_file}'"])
    mito.code_options_update({'as_function': True, 'import_custom_python_code': False, 'call_function': True, 'function_name': 'function', 'function_params': {}})

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        'import pandas as pd',
        '',
        'def function(txt_path):',
        f"{TAB}# Read in filepaths as dataframes",
        f"{TAB}txt = pd.read_csv(txt_path)",
        f'{TAB}',
        f"{TAB}return txt",
        "",
        "txt = function('txt.csv')"
    ]

    os.remove(tmp_file)

def test_transpile_as_function_string_params_no_args_update():
    tmp_file = 'txt.csv'
    df1 = pd.DataFrame({'A': [1], 'B': [2]})
    df1.to_csv(tmp_file, index=False)

    mito = create_mito_wrapper(str(tmp_file))
    mito.code_options_update({'as_function': True, 'import_custom_python_code': False, 'call_function': True, 'function_name': 'function', 'function_params': {}})

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        'import pandas as pd',
        '',
        'def function(txt_path):',
        f"{TAB}# Read in filepaths as dataframes",
        f"{TAB}txt = pd.read_csv(txt_path)",
        f'{TAB}',
        f"{TAB}return txt",
        "",
        'txt = function("txt.csv")'
    ]

    os.remove(tmp_file)

def test_transpile_as_function_both_params():
    tmp_file = 'txt.csv'
    df1 = pd.DataFrame({'A': [1], 'B': [2]})
    df1.to_csv(tmp_file, index=False)

    mito = create_mito_wrapper(df1, str(tmp_file), arg_names=['df1', f"'{tmp_file}'"])
    mito.code_options_update({'as_function': True, 'import_custom_python_code': False, 'call_function': True, 'function_name': 'function', 'function_params': {}})

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        'import pandas as pd',
        '',
        'def function(df1, txt_path):',
        f"{TAB}# Read in filepaths as dataframes",
        f"{TAB}txt = pd.read_csv(txt_path)",
        f'{TAB}',
        f"{TAB}return df1, txt",
        "",
        "df1, txt = function(df1, 'txt.csv')"
    ]

    os.remove(tmp_file)


def test_transpile_pivot_table_indents():
    df1 = pd.DataFrame(data={'Name': ['Nate', 'Nate'], 'Height': [4, 5]})
    mito = create_mito_wrapper(df1, arg_names=['df1'])

    mito.code_options_update({'as_function': True, 'import_custom_python_code': False, 'call_function': True, 'function_name': 'function', 'function_params': {}})

    mito.pivot_sheet(
        0, 
        ['Name'],
        [],
        {'Height': ['sum']}
    )
    
    print("\n".join(mito.transpiled_code))
    assert "    pivot_table = tmp_df.pivot_table(\n        index=['Name'],\n        values=['Height'],\n        aggfunc={'Height': ['sum']}\n    )" in mito.transpiled_code


def test_transpile_as_function_single_param(tmp_path):
    tmp_file = str(tmp_path / 'txt.csv')
    df1 = pd.DataFrame({'A': [1], 'B': [2]})
    df1.to_csv(tmp_file, index=False)

    mito = create_mito_wrapper()
    mito.simple_import([tmp_file])
    mito.code_options_update({'as_function': True, 'import_custom_python_code': False, 'call_function': True, 'function_name': 'function', 'function_params': {'var_name': f"r'{tmp_file}'"}})

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        "import pandas as pd",
        "",
        "def function(var_name):",
        f"{TAB}txt = pd.read_csv(var_name)",
        f'{TAB}',
        f"{TAB}return txt",
        "",
        f"var_name = r'{tmp_file}'",
        "",
        f"txt = function(var_name)"
    ]

def test_transpile_as_function_single_param_no_call(tmp_path):
    tmp_file = str(tmp_path / 'txt.csv')
    df1 = pd.DataFrame({'A': [1], 'B': [2]})
    df1.to_csv(tmp_file, index=False)

    mito = create_mito_wrapper()
    mito.simple_import([tmp_file])
    mito.code_options_update_no_check_transpiled({'as_function': True, 'import_custom_python_code': False, 'call_function': False, 'function_name': 'function', 'function_params': {'var_name': f"r'{tmp_file}'"}})

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        "import pandas as pd",
        "",
        "def function(var_name):",
        f"{TAB}txt = pd.read_csv(var_name)",
        f'{TAB}',
        f"{TAB}return txt",
        "",
    ]


def test_transpile_as_function_both_params_and_additional():
    tmp_file = 'txt.csv'
    df1 = pd.DataFrame({'A': [1], 'B': [2]})
    df1.to_csv(tmp_file, index=False)

    mito = create_mito_wrapper(df1, str(tmp_file), arg_names=['df1', f"'{tmp_file}'"])
    mito.simple_import([tmp_file])
    mito.code_options_update({'as_function': True, 'import_custom_python_code': False, 'call_function': True, 'function_name': 'function', 'function_params': {'var_name': f"r'{tmp_file}'"}})

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        'import pandas as pd',
        '',
        'def function(df1, txt_path, var_name):',
        f"{TAB}# Read in filepaths as dataframes",
        f"{TAB}txt = pd.read_csv(txt_path)",
        f'{TAB}',
        '    txt_1 = pd.read_csv(var_name)',
        f'{TAB}',
        f"{TAB}return df1, txt, txt_1",
        "",
        f"var_name = r'{tmp_file}'",
        "",
        f"df1, txt, txt_1 = function(df1, 'txt.csv', var_name)"
    ]

    os.remove(tmp_file)

def test_transpile_as_function_single_param_multiple_times(tmp_path):
    tmp_file = str(tmp_path / 'txt.csv')
    df1 = pd.DataFrame({'A': [1], 'B': [2]})
    df1.to_csv(tmp_file, index=False)

    mito = create_mito_wrapper()
    mito.simple_import([tmp_file])
    mito.simple_import([tmp_file])
    mito.code_options_update({'as_function': True, 'import_custom_python_code': False, 'call_function': True, 'function_name': 'function', 'function_params': {'var_name': f"r'{tmp_file}'"}})

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        "import pandas as pd",
        "",
        "def function(var_name):",
        f"{TAB}txt = pd.read_csv(var_name)",
        f"{TAB}txt_1 = pd.read_csv(var_name)",
        f'{TAB}',
        f"{TAB}return txt, txt_1",
        "",
        f"var_name = r'{tmp_file}'",
        "",
        f"txt, txt_1 = function(var_name)"
    ]

@pandas_post_1_2_only
@python_post_3_6_only
@pytest.mark.parametrize("df_names, expected_in_transpile, expected_in_function", [(None, 'df1', 'import_dataframe_0'), (['test_df_name'], 'test_df_name', 'import_dataframe_0')])
def test_transpile_fully_parameterized_function_string_no_df_names(tmp_path, df_names, expected_in_transpile, expected_in_function):
    tmp_file1 = str(tmp_path / 'txt.csv')
    tmp_file2 = str(tmp_path / 'file.xlsx')
    tmp_exportfile1 = str(tmp_path / 'export.csv')
    tmp_exportfile2 = str(tmp_path / 'export.xlsx')
    df1 = pd.DataFrame({'A': [1], 'B': [2]})
    df1.to_csv(tmp_file1, index=False)
    df1.to_excel(tmp_file2, index=False)

    mito = create_mito_wrapper(df1, arg_names=df_names)
    # Test imports for excel and CSV
    mito.simple_import([tmp_file1])
    mito.excel_import(tmp_file2, sheet_names=['Sheet1'], has_headers=True, skiprows=0)

    # Test exports for excel and CSV
    mito.export_to_file('csv', [0], tmp_exportfile1)
    mito.export_to_file('excel', [1], tmp_exportfile2)

    mito.code_options_update({'as_function': False, 'import_custom_python_code': False, 'call_function': False, 'function_name': 'function', 'function_params': {}})
    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        "import pandas as pd",
        "",
        f"txt = pd.read_csv(r'{tmp_file1}')",
        "",
        f"sheet_df_dictonary = pd.read_excel(r'{tmp_file2}', engine='openpyxl', sheet_name=['Sheet1'], skiprows=0)",
        "Sheet1 = sheet_df_dictonary['Sheet1']",
        "",
        f"{expected_in_transpile}.to_csv(r'{tmp_exportfile1}', "
        'index=False)',
        '',
        'with '
        f"pd.ExcelWriter(r'{tmp_exportfile2}', "
        'engine="openpyxl") as writer:',
        '    txt.to_excel(writer, sheet_name="txt", index=False)',
        '',
    ]

    assert mito.mito_backend.fully_parameterized_function == f"""from mitosheet.public.v3 import *
import pandas as pd

def function({expected_in_function}, file_name_import_csv_0, file_name_import_excel_0, file_name_export_csv_0, file_name_export_excel_0):
    txt = pd.read_csv(file_name_import_csv_0)
    
    sheet_df_dictonary = pd.read_excel(file_name_import_excel_0, engine='openpyxl', sheet_name=['Sheet1'], skiprows=0)
    Sheet1 = sheet_df_dictonary['Sheet1']
    
    {expected_in_function}.to_csv(file_name_export_csv_0, index=False)
    
    with pd.ExcelWriter(file_name_export_excel_0, engine="openpyxl") as writer:
        txt.to_excel(writer, sheet_name="txt", index=False)
    
    return {expected_in_function}, txt, Sheet1
"""
    assert mito.mito_backend.param_metadata == [
        {
            'original_value': 'test_df_name' if df_names else 'df1',
            'type': 'import',
            'subtype': 'import_dataframe',
            'required': True,
            'name': expected_in_function
        },
        {
            'original_value': tmp_file1,
            'type': 'import',
            'subtype': 'file_name_import_csv',
            'name': 'file_name_import_csv_0',
            'required': False
        },
        {
            'original_value': tmp_file2,
            'type': 'import',
            'subtype': 'file_name_import_excel',
            'name': 'file_name_import_excel_0',
            'required': False
        },
        {
            'original_value': tmp_exportfile1,
            'type': 'export',
            'subtype': 'file_name_export_csv',
            'name': 'file_name_export_csv_0',
            'required': False
        },
        {
            'original_value': tmp_exportfile2,
            'type': 'export',
            'subtype': 'file_name_export_excel',
            'name': 'file_name_export_excel_0',
            'required': False
        }
    ]


@pandas_post_1_2_only
@python_post_3_6_only
def test_transpile_fully_parameterized_function_string_no_df_name_param(tmp_path):
    tmp_file1 = str(tmp_path / 'txt.csv')
    tmp_file2 = str(tmp_path / 'file.xlsx')
    tmp_exportfile1 = str(tmp_path / 'export.csv')
    tmp_exportfile2 = str(tmp_path / 'export.xlsx')
    df1 = pd.DataFrame({'A': [1], 'B': [2]})
    df1.to_csv(tmp_file1, index=False)
    df1.to_excel(tmp_file2, index=False)

    mito = create_mito_wrapper()
    # Test imports for excel and CSV
    mito.simple_import([tmp_file1])
    mito.excel_import(tmp_file2, sheet_names=['Sheet1'], has_headers=True, skiprows=0)

    # Test exports for excel and CSV
    mito.export_to_file('csv', [0], tmp_exportfile1)
    mito.export_to_file('excel', [1], tmp_exportfile2)

    mito.code_options_update({'as_function': False, 'import_custom_python_code': False, 'call_function': False, 'function_name': 'function', 'function_params': {}})
    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        "import pandas as pd",
        "",
        f"txt = pd.read_csv(r'{tmp_file1}')",
        "",
        f"sheet_df_dictonary = "
        f"pd.read_excel(r'{tmp_file2}', "
        "engine='openpyxl', sheet_name=['Sheet1'], skiprows=0)",
        "Sheet1 = sheet_df_dictonary['Sheet1']",
        "",
        f"txt.to_csv(r'{tmp_exportfile1}', "
        'index=False)',
        '',
        'with '
        f"pd.ExcelWriter(r'{tmp_exportfile2}', "
        'engine="openpyxl") as writer:',
        '    Sheet1.to_excel(writer, sheet_name="Sheet1", index=False)',
        '',
    ]

    assert mito.mito_backend.fully_parameterized_function == f"""from mitosheet.public.v3 import *
import pandas as pd

def function(file_name_import_csv_0, file_name_import_excel_0, file_name_export_csv_0, file_name_export_excel_0):
    txt = pd.read_csv(file_name_import_csv_0)
    
    sheet_df_dictonary = pd.read_excel(file_name_import_excel_0, engine='openpyxl', sheet_name=['Sheet1'], skiprows=0)
    Sheet1 = sheet_df_dictonary['Sheet1']
    
    txt.to_csv(file_name_export_csv_0, index=False)
    
    with pd.ExcelWriter(file_name_export_excel_0, engine="openpyxl") as writer:
        Sheet1.to_excel(writer, sheet_name="Sheet1", index=False)
    
    return txt, Sheet1
"""
    assert mito.mito_backend.param_metadata == [
        {
            'original_value': tmp_file1,
            'type': 'import',
            'subtype': 'file_name_import_csv',
            'name': 'file_name_import_csv_0',
            'required': False
        },
        {
            'original_value': tmp_file2,
            'type': 'import',
            'subtype': 'file_name_import_excel',
            'name': 'file_name_import_excel_0',
            'required': False
        },
        {
            'original_value': tmp_exportfile1,
            'type': 'export',
            'subtype': 'file_name_export_csv',
            'name': 'file_name_export_csv_0',
            'required': False
        },
        {
            'original_value': tmp_exportfile2,
            'type': 'export',
            'subtype': 'file_name_export_excel',
            'name': 'file_name_export_excel_0',
            'required': False
        }
    ]


@pandas_post_1_2_only
@python_post_3_6_only
def test_fully_parameterized_function_custom_imports():
    mito = create_mito_wrapper(sheet_functions=[ADDONE], importers=[custom_import])
    mito.user_defined_import('custom_import', {})
    mito.set_formula('=ADDONE(A)', 0, 'B', add_column=True)

    mito.code_options_update({'as_function': False, 'import_custom_python_code': False, 'call_function': False, 'function_name': 'function', 'function_params': {}})
    assert "\n".join(mito.transpiled_code) == """from mitosheet.public.v3 import *

df0 = custom_import()

df0['B'] = ADDONE(df0['A'])
"""

    assert mito.mito_backend.fully_parameterized_function == """from mitosheet.public.v3 import *
from mitosheet.tests.test_transpile import custom_import, ADDONE

def function():
    df0 = custom_import()
    
    df0['B'] = ADDONE(df0['A'])
    
    return df0
"""

def test_transpile_as_function_multiple_params(tmp_path):
    tmp_file1 = str(tmp_path / 'txt.csv')
    tmp_file2 = str(tmp_path / 'file.csv')
    df1 = pd.DataFrame({'A': [1], 'B': [2]})
    df1.to_csv(tmp_file1, index=False)
    df1.to_csv(tmp_file2, index=False)

    mito = create_mito_wrapper()
    mito.simple_import([tmp_file1])
    mito.simple_import([tmp_file2])
    mito.code_options_update({'as_function': True, 'import_custom_python_code': False, 'call_function': True, 'function_name': 'function', 'function_params': {'var_name1': f"r'{tmp_file1}'", 'var_name2': f"r'{tmp_file2}'"}})

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        "import pandas as pd",
        "",
        "def function(var_name1, var_name2):",
        f"{TAB}txt = pd.read_csv(var_name1)",
        f"{TAB}file = pd.read_csv(var_name2)",
        f'{TAB}',
        f"{TAB}return txt, file",
        "",
        f"var_name1 = r'{tmp_file1}'",
        f"var_name2 = r'{tmp_file2}'",
        "",
        f"txt, file = function(var_name1, var_name2)"
    ]

@pandas_post_1_2_only
@python_post_3_6_only
def test_transpile_parameterize_excel_imports(tmp_path):
    tmp_file = str(tmp_path / 'txt.xlsx')
    df1 = pd.DataFrame({'A': [1], 'B': [2]})
    df1.to_excel(tmp_file, index=False)

    mito = create_mito_wrapper()
    mito.excel_import(tmp_file, sheet_names=['Sheet1'], has_headers=True, skiprows=0)
    mito.excel_range_import(tmp_file, {'type': 'sheet name', 'value': 'Sheet1'}, [{'type': 'range', 'df_name': 'dataframe_1', 'value': 'A1:B2'}], convert_csv_to_xlsx=False)
    mito.code_options_update({'as_function': True, 'import_custom_python_code': False, 'call_function': True, 'function_name': 'function', 'function_params': {'var_name': f"r'{tmp_file}'"}})

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        "import pandas as pd",
        "",
        "def function(var_name):",
        f"{TAB}sheet_df_dictonary = pd.read_excel(var_name, engine='openpyxl', sheet_name=['Sheet1'], skiprows=0)",
        f"{TAB}Sheet1 = sheet_df_dictonary['Sheet1']",
        f'{TAB}',
        f"{TAB}dataframe_1 = pd.read_excel(var_name, sheet_name='Sheet1', skiprows=0, nrows=1, usecols='A:B')",
        f'{TAB}',
        f"{TAB}return Sheet1, dataframe_1",
        "",
        f"var_name = r'{tmp_file}'",
        "",
        f"Sheet1, dataframe_1 = function(var_name)"
    ]

def test_transpile_with_function_params_over_mitosheet():
    df1 = pd.DataFrame({'A': [1], 'B': [2]})
    df2 = pd.DataFrame({'A': [1], 'B': [2]})
    mito = create_mito_wrapper(df1, df2, arg_names=['df', 'df_copy'])
    mito.add_column(0, 'C')
    mito.add_column(1, 'C')

    mito.code_options_update({'as_function': True, 'import_custom_python_code': False, 'call_function': True, 'function_name': 'function', 'function_params': {'param': "df"}})

    
    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        "",
        "def function(param, df_copy):",
        f"{TAB}param['C'] = 0",
        f"{TAB}",
        f"{TAB}df_copy['C'] = 0",
        f"{TAB}",
        f"{TAB}return param, df_copy",
        "",
        f"param = df",
        "",
        f"param, df_copy = function(param, df_copy)"
    ]

def test_transpile_does_not_effect_chars_in_strings():
    mito = create_mito_wrapper()
    quote = '"'
    mito.ai_transformation(
        'do a test',
        'v1',
        'test',
        'test',
        """
df = pd.DataFrame({'A': ["has a new \
line in it", '\t', '     ']})
        """
    )

    mito.code_options_update({'as_function': True, 'import_custom_python_code': False, 'call_function': True, 'function_name': 'function', 'function_params': {}})

    assert "\n".join(mito.transpiled_code) == """from mitosheet.public.v3 import *
import pandas as pd

def function():
    df = pd.DataFrame({'A': ["has a new \
line in it", '\t', '     ']})
    
    return df

df = function()"""

def test_transpile_with_multiline_ai_completion():
    mito = create_mito_wrapper()
    mito.ai_transformation(
        'do a test',
        'v1',
        'test',
        'test',
        """
import pandas as pd

# create sample dataframe
df = pd.DataFrame({'A': [1, 2, 3], 'B': [4, 5, 6], 'C': [7, 8, 9]})

print(df)
        """
    )

    mito.code_options_update({'as_function': True, 'import_custom_python_code': False, 'call_function': True, 'function_name': 'function', 'function_params': {}})

    assert "\n".join(mito.transpiled_code) == """from mitosheet.public.v3 import *

def function():
    import pandas as pd
    
    # create sample dataframe
    df = pd.DataFrame({'A': [1, 2, 3], 'B': [4, 5, 6], 'C': [7, 8, 9]})
    
    print(df)
    
    return df

df = function()"""

def test_transpiled_with_export_to_csv_singular():
    df = pd.DataFrame({'A': [1, 2, 3]})
    mito = create_mito_wrapper(df, arg_names=['df'])
    mito.export_to_file('csv', [0], 'te"st.csv')

    assert [('df', 'import', 'import_dataframe'), ("r'te" + '"' + "st.csv'", 'export', 'file_name_export_csv')] == get_parameterizable_params({}, mito.mito_backend.steps_manager)

    mito.code_options_update({'as_function': True, 'import_custom_python_code': False, 'call_function': True, 'function_name': 'function', 'function_params': {'path': "r'te" + '"' + "st.csv'"}})

    assert "\n".join(mito.transpiled_code) == """from mitosheet.public.v3 import *

def function(df, path):
    df.to_csv(path, index=False)
    
    return df

path = r'te"st.csv'

df = function(df, path)"""

def test_transpiled_with_export_to_csv_multiple():
    df = pd.DataFrame({'A': [1, 2, 3]})
    mito = create_mito_wrapper(df, df, arg_names=['df1', 'df2'])
    mito.export_to_file('csv', [0, 1], 'test.csv')

    assert [('df1', 'import', 'import_dataframe'), ('df2', 'import', 'import_dataframe'), ("r'test_0.csv'", 'export', 'file_name_export_csv'), ("r'test_1.csv'", 'export', 'file_name_export_csv')] == get_parameterizable_params({}, mito.mito_backend.steps_manager)

    mito.code_options_update({'as_function': True, 'import_custom_python_code': False, 'call_function': True, 'function_name': 'function', 'function_params': {'path_0': "r'test_0.csv'", 'path_1': "r'test_1.csv'"}})

    assert "\n".join(mito.transpiled_code) == """from mitosheet.public.v3 import *

def function(df1, df2, path_0, path_1):
    df1.to_csv(path_0, index=False)
    df2.to_csv(path_1, index=False)
    
    return df1, df2

path_0 = r'test_0.csv'
path_1 = r'test_1.csv'

df1, df2 = function(df1, df2, path_0, path_1)"""

def test_transpiled_with_export_to_xlsx_single():
    df = pd.DataFrame({'A': [1, 2, 3]})
    mito = create_mito_wrapper(df, arg_names=['df'])
    mito.export_to_file('excel', [0], "te'st.xlsx")

    assert [('df', 'import', 'import_dataframe'), ('r"te' + "'" + 'st.xlsx"', 'export', 'file_name_export_excel')] == get_parameterizable_params({}, mito.mito_backend.steps_manager)

    mito.code_options_update({'as_function': True, 'import_custom_python_code': False, 'call_function': True, 'function_name': 'function', 'function_params': {'path': 'r"te' + "'" + 'st.xlsx"'}})

    assert "\n".join(mito.transpiled_code) == """from mitosheet.public.v3 import *
import pandas as pd

def function(df, path):
    with pd.ExcelWriter(path, engine="openpyxl") as writer:
        df.to_excel(writer, sheet_name="df", index=False)
    
    return df

path = r"te'st.xlsx"

df = function(df, path)"""

def test_transpiled_with_export_to_xlsx_multiple():
    df = pd.DataFrame({'A': [1, 2, 3]})
    mito = create_mito_wrapper(df, df, arg_names=['df1', 'df2'])
    mito.export_to_file('excel', [0, 1], 'test.xlsx')

    assert [('df1', 'import', 'import_dataframe'), ('df2', 'import', 'import_dataframe'), ("r'test.xlsx'", 'export', 'file_name_export_excel')] == get_parameterizable_params({}, mito.mito_backend.steps_manager)

    mito.code_options_update({'as_function': True, 'import_custom_python_code': False, 'call_function': True, 'function_name': 'function', 'function_params': {'path_0': "r'test.xlsx'"}})

    assert "\n".join(mito.transpiled_code) == """from mitosheet.public.v3 import *
import pandas as pd

def function(df1, df2, path_0):
    with pd.ExcelWriter(path_0, engine="openpyxl") as writer:
        df1.to_excel(writer, sheet_name="df1", index=False)
        df2.to_excel(writer, sheet_name="df2", index=False)
    
    return df1, df2

path_0 = r'test.xlsx'

df1, df2 = function(df1, df2, path_0)"""


def test_code_options_pass_param_subtype():
    df = pd.DataFrame({'A': [1, 2, 3]})
    mito = create_mito_wrapper(df, arg_names=['df'])
    mito.export_to_file('csv', [0], 'test.csv')

    mito.code_options_update({'as_function': True, 'import_custom_python_code': False, 'call_function': True, 'function_name': 'function', 'function_params': 'file_name_export_csv'})

    assert "\n".join(mito.transpiled_code) == """from mitosheet.public.v3 import *

def function(df, file_name_export_csv_0):
    df.to_csv(file_name_export_csv_0, index=False)
    
    return df

file_name_export_csv_0 = r'test.csv'

df = function(df, file_name_export_csv_0)"""

def test_code_options_pass_param_subtype_multiple_of_subtype():
    df = pd.DataFrame({'A': [1, 2, 3]})
    mito = create_mito_wrapper(df, arg_names=['df'])
    mito.export_to_file('csv', [0], 'test.csv')
    mito.export_to_file('csv', [0], 'test1.csv')

    mito.code_options_update({'as_function': True, 'import_custom_python_code': False, 'call_function': True, 'function_name': 'function', 'function_params': 'file_name_export_csv'})

    assert "\n".join(mito.transpiled_code) == """from mitosheet.public.v3 import *

def function(df, file_name_export_csv_0, file_name_export_csv_1):
    df.to_csv(file_name_export_csv_0, index=False)
    
    df.to_csv(file_name_export_csv_1, index=False)
    
    return df

file_name_export_csv_0 = r'test.csv'
file_name_export_csv_1 = r'test1.csv'

df = function(df, file_name_export_csv_0, file_name_export_csv_1)"""

def test_code_options_pass_multiple_param_subtype():
    df = pd.DataFrame({'A': [1, 2, 3]})
    mito = create_mito_wrapper(df, arg_names=['df'])
    mito.export_to_file('csv', [0], 'test.csv')
    mito.export_to_file('excel', [0], 'test.xlsx')

    mito.code_options_update({'as_function': True, 'import_custom_python_code': False, 'call_function': True, 'function_name': 'function', 'function_params': ['file_name_export_csv', 'file_name_export_excel']})

    assert "\n".join(mito.transpiled_code) == """from mitosheet.public.v3 import *
import pandas as pd

def function(df, file_name_export_csv_0, file_name_export_excel_0):
    df.to_csv(file_name_export_csv_0, index=False)
    
    with pd.ExcelWriter(file_name_export_excel_0, engine="openpyxl") as writer:
        df.to_excel(writer, sheet_name="df", index=False)
    
    return df

file_name_export_csv_0 = r'test.csv'
file_name_export_excel_0 = r'test.xlsx'

df = function(df, file_name_export_csv_0, file_name_export_excel_0)"""

def ADDONE(x):
    return x + 1

def custom_import():
    return pd.DataFrame({'A': [1, 2, 3]})

def custom_edit(df: pd.DataFrame) -> pd.DataFrame:
    return df + 1

def test_code_options_include_functions():
    
    mito = create_mito_wrapper(sheet_functions=[ADDONE], importers=[custom_import], editors=[custom_edit])
    mito.user_defined_import('custom_import', {})
    mito.set_formula('=ADDONE(A)', 0, 'B', add_column=True)
    mito.user_defined_edit('custom_edit', {'df': 'df0'})

    mito.code_options_update({'as_function': True, 'import_custom_python_code': True, 'call_function': True, 'function_name': 'function', 'function_params': {}})
    print(mito.transpiled_code)
    assert "\n".join(mito.transpiled_code) == """from mitosheet.public.v3 import *
from mitosheet.tests.test_transpile import custom_import, ADDONE, custom_edit

def function():
    df0 = custom_import()
    
    df0['B'] = ADDONE(df0['A'])
    
    df0 = custom_edit(df=df0)
    
    return df0

df0 = function()"""


def test_transpile_handles_new_line_character():
    df = pd.DataFrame({'A\nB': [1,2,3]})
    mito = create_mito_wrapper(df)

    mito.filter(0, 'A\nB', "And", FC_NUMBER_GREATER, 1)

    assert "\n".join(mito.transpiled_code_with_comments) == """from mitosheet.public.v3 import *

# Filtered A
# B
df1 = df1[df1['A\\nB'] > 1]
"""


def test_transpile_optimizes_renames_out_of_order():
    df = pd.DataFrame({'A': [1], 'B': [2]})
    mito = create_mito_wrapper(df, df)

    mito.rename_column(0, 'A', 'AA')
    mito.rename_column(1, 'A', 'AA')
    mito.rename_column(0, 'B', 'BB')

    assert mito.dfs[0].equals(pd.DataFrame({'AA': [1], 'BB': [2]}))
    assert mito.dfs[1].equals(pd.DataFrame({'AA': [1], 'B': [2]}))

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *', 
        '', 
        "df1.rename(columns={'A': 'AA', 'B': 'BB'}, inplace=True)", 
        '', 
        "df2.rename(columns={'A': 'AA'}, inplace=True)", 
        '',
    ]

def test_transpile_optimizes_many_renames_out_of_order():
    df = pd.DataFrame({'A': [1], 'B': [2], 'C': [3]})
    mito = create_mito_wrapper(df, df)

    mito.rename_column(0, 'A', 'AA')
    mito.rename_column(1, 'A', 'AA')
    mito.rename_column(0, 'B', 'BB')
    mito.rename_column(1, 'B', 'BB')
    mito.rename_column(0, 'C', 'CC')
    mito.rename_column(1, 'C', 'CC')

    assert mito.dfs[0].equals(pd.DataFrame({'AA': [1], 'BB': [2], 'CC': [3]}))
    assert mito.dfs[1].equals(pd.DataFrame({'AA': [1], 'BB': [2], 'CC': [3]}))

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *', 
        '', 
        "df1.rename(columns={'A': 'AA', 'B': 'BB', 'C': 'CC'}, inplace=True)", 
        '', 
        "df2.rename(columns={'A': 'AA', 'B': 'BB', 'C': 'CC'}, inplace=True)", 
        '',
    ]

def test_can_optimize_around_imports(tmp_path):
    df = pd.DataFrame({'A': [1], 'B': [2], 'C': [3]})
    path = str(tmp_path / 'test.csv')
    df.to_csv(path, index=False)

    mito = create_mito_wrapper()
    mito.simple_import([path])
    mito.add_column(0, 'D')
    mito.add_column(0, 'E')
    mito.simple_import([path])
    mito.delete_columns(0, ['D', 'E'])

    assert len(mito.dfs) == 2
    assert mito.dfs[0].equals(df)
    assert mito.dfs[1].equals(df)

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *', 
        'import pandas as pd', 
        '', 
        f"test = pd.read_csv(r'{path}')", 
        f"test_1 = pd.read_csv(r'{path}')", 
        '', 
    ]

def test_can_optimize_imports_together_if_just_edits(tmp_path):
    df = pd.DataFrame({'A': [1], 'B': [2], 'C': [3]})
    new_df = pd.DataFrame({'A': [1], 'B': [2], 'C': [3], 'D': [0], 'E': [0]})

    path = str(tmp_path / 'test.csv')
    df.to_csv(path, index=False)

    mito = create_mito_wrapper()
    mito.simple_import([path])
    mito.add_column(0, 'D')
    mito.add_column(0, 'E')
    mito.simple_import([path])

    assert len(mito.dfs) == 2
    assert mito.dfs[0].equals(new_df)
    assert mito.dfs[1].equals(df)

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *', 
        'import pandas as pd', 
        '', 
        f"test = pd.read_csv(r'{path}')", 
        f"test_1 = pd.read_csv(r'{path}')", 
        '', 
        "test['D'] = 0", 
        '', 
        "test['E'] = 0", 
        ''
    ]


def test_edit_dataframe_after_full_clear_does_not_reorder_strangely(tmp_path):
    df = pd.DataFrame({'A': [1], 'B': [2], 'C': [3]})
    new_df = pd.DataFrame({'A': [1], 'B': [2], 'C': [3], 'D': [0]})

    path = str(tmp_path / 'test.csv')
    df.to_csv(path, index=False)

    mito = create_mito_wrapper()
    mito.simple_import([path])
    mito.delete_dataframe(0)
    mito.simple_import([path])
    mito.add_column(0, 'D')

    assert len(mito.dfs) == 1
    assert mito.dfs[0].equals(new_df)
    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *', 
        'import pandas as pd', 
        '', 
        f"test = pd.read_csv(r'{path}')", 
        '', 
        "test['D'] = 0", 
        ''
    ]


def test_all_edits_optimized_after_simple_imports(tmp_path):
    df = pd.DataFrame({'A': [1], 'B': [2], 'C': [3], 'D': [0], 'E': [0]})
    path = str(tmp_path / 'test.csv')
    df.to_csv(path, index=False)
    
    mito = create_mito_wrapper()
    mito.simple_import([path])
    mito.simple_import([path])

    mito.add_column(0, 'F')
    mito.add_column(1, 'F')

    mito.set_formula('=A + 1', 0, 'F')
    mito.set_formula('=B + 1', 1, 'F')

    assert len(mito.dfs) == 2
    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *', 
        'import pandas as pd', 
        '', 
        f"test = pd.read_csv(r'{path}')",
        f"test_1 = pd.read_csv(r'{path}')", 
        '', 
        "test['F'] = test['A'] + 1",
        '',
        "test_1['F'] = test_1['B'] + 1",
        ''
    ]

@pytest.mark.skip(reason="This test is non-deterministic. Sometimes tmp_df = test[[\'B\', \'A\']].copy(), sometimes tmp_df = test[[\'A\', \'B\']].copy() ")
def test_all_edits_optimized_after_import_pivot(tmp_path):
    df = pd.DataFrame({'A': [1], 'B': [2], 'C': [3], 'D': [0], 'E': [0]})
    path = str(tmp_path / 'test.csv')
    df.to_csv(path, index=False)
    
    mito = create_mito_wrapper()
    mito.simple_import([path])

    mito.pivot_sheet(
        0, 
        ['A'],
        [],
        {'B': ['sum']}
    )

    mito.delete_columns(0, ['A'])
    mito.add_column(1, 'NEW')
    mito.delete_columns(0, ['B'])
    mito.set_formula('=A + 1', 1, 'NEW')
    mito.delete_columns(0, ['C'])

    assert len(mito.dfs) == 2
    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *', 
        'import pandas as pd', 
        '', 
        f"test = pd.read_csv(r'{path}')",
        '', 
        'tmp_df = test[[\'B\', \'A\']].copy()',
        'pivot_table = tmp_df.pivot_table(\n    index=[\'A\'],\n    values=[\'B\'],\n    aggfunc={\'B\': [\'sum\']}\n)',
        'pivot_table = pivot_table.set_axis([flatten_column_header(col) for col in pivot_table.keys()], axis=1)',
        'test_pivot = pivot_table.reset_index()',
        '',
        'test.drop([\'A\', \'B\', \'C\'], axis=1, inplace=True)',
        '',
        "test_pivot['NEW'] = test_pivot['A'] + 1",
        '',
    ]

def test_does_not_reorder_around_empty_code_chunk():
    df = pd.DataFrame({'A': [1], 'B': [2], 'C': [3]})
    mito = create_mito_wrapper(df)

    mito.add_column(0, 'D')
    mito.delete_columns(0, ['A', 'B'])
    result = mito.generate_graph('test', BAR, 0, False, ['C'], [], '400', '400')
    assert result
    assert mito.dfs[0].equals(pd.DataFrame({'C': [3], 'D': [0]}))


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from copy import copy
from typing import Any, Dict, List, Optional
from mitosheet.saved_analyses.step_upgraders.utils_column_header_to_column_id import \
    replace_headers_with_id


def replace_column_header_in_params(
        old_column_header: str, 
        new_column_header: str, 
        params: Dict[str, Any],
        keys: Optional[List[str]]=None
    ) -> Dict[str, Any]:
    """
    Replaces old_column_header with new_column_header wherever it occurs in params.

    If you only want to change some keys in the params, pass them as a list, and 
    only these params will be searched through. Useful for making sure you don't 
    change, for example, merge keys two when you only want to change merge key one.
    """
    if keys is None:
        keys = list(params.keys())
    
    new_params = copy(params)
    for key in keys:
        value = params[key]

        new_value = value
        if isinstance(value, str):
            if value == new_column_header:
                new_value = old_column_header
        elif isinstance(value, list):
            for idx, element in enumerate(value):
                if element == new_column_header:
                    value[idx] = old_column_header
        elif isinstance(value, dict):
            new_value = dict()
            for subkey, subvalue in value.items():
                if subkey == new_column_header:
                    subkey = old_column_header
                new_value[subkey] = subvalue
        
        new_params[key] = new_value
    
    return new_params

def change_column_header_in_all_future_steps(
        sheet_index: int,
        old_column_header: str, 
        new_column_header: str, 
        later_steps: List[Dict[str, Any]]
    ) -> List[Dict[str, Any]]:
    """
    This helper function changes all future steps in the analysis
    after the rename by switching new_column_header to old_column_header
    wherever it finds it. This effectively undoes the rename in the
    rest of the analysis so it still generates IDs correctly.

    This, in turns, means that when these steps are upgraded to use column
    ids instead of column headers, they will reference the correct id.  
    """
    for step in later_steps:
        params = step['params']
        new_params = params

        # If a sheet_index is the params, we only modify the step if the sheet index
        # being modified is the same as the sheet index where the column header was renamed
        if 'sheet_index' in params:
            if params['sheet_index'] != sheet_index:
                continue

        # Instead of handling each step one by one, we just try and upgrade them
        # all at once, taking special care to not upgrade merge in an invalid way.
        # Merge is the only step we have to be careful about, as it is the only
        # step that touches two dataframes
        if step['step_type'] == 'merge':
            if params['sheet_index_one'] == sheet_index:
                new_params = replace_column_header_in_params(
                    old_column_header,
                    new_column_header,
                    new_params,
                    keys=['merge_key_one', 'selected_columns_one']
                )
            if params['sheet_index_two'] == sheet_index:
                new_params = replace_column_header_in_params(
                    old_column_header,
                    new_column_header,
                    new_params,
                    keys=['merge_key_two', 'selected_columns_two']
                )
        else:
            new_params = replace_column_header_in_params(
                old_column_header,
                new_column_header,
                new_params
            )

        step['params'] = new_params
    
    return later_steps


def upgrade_rename_column_1_to_2(step: Dict[str, Any], later_steps: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """
    Moves to using column id instead of column header.

    OLD: {
        'step_version': 1, 
        'step_type': "rename_column",
        'params': {
            sheet_index: 0,
            old_column_header: _header_,
            new_column_header: "New value"
        }
    }

    NEW: {
        'step_version': 2, 
        'step_type': "rename_column",
        'params': {
            sheet_index: 0,
            column_id: _id_,
            new_column_header: "New value"
        }
    }

    But this annoying function is also responsible for making the move from column header
    to column ID across the codebase work properly. 

    Here's what broke:
    1. You have an analysis, where you create a column, rename it, and then edit it.
       [Create A, Rename A -> B, Edit B]
    2. The step upgrading alone just statically moves from column header to column id,
       and so you end up with [Create ID_A, Rename ID_A -> B, Edit ID_B]
    3. So, before we do the stpe upgrade, we have to go through the analysis, find _all_
       places where renames occured, and fix them up to be the original column header
       names. 
    
    NOTE: this does not do the step upgrading itself. It will just turn the original 
    steps data into [Create A, Rename A -> B, Edit A]. Thus, the step upgraders will
    end up with: [Create ID_A, Rename ID_A -> B, Edit ID_A].
    """    
    params = step['params']

    # First, we change the column header in all future steps
    later_steps = change_column_header_in_all_future_steps(
        params['sheet_index'],
        params['old_column_header'],
        params['new_column_header'],
        later_steps
    )

    # Then, switch to the column id
    params = replace_headers_with_id(params, 'old_column_header', 'column_id')

    return [{
        "step_version": 2, 
        "step_type": "rename_column", 
        "params": params
    }] + later_steps




#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

import os
from time import perf_counter
from typing import Any, Dict, List, Optional, Set, Tuple

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.export_to_file_code_chunk import \
    ExportToFileCodeChunk
from mitosheet.excel_utils import get_df_name_as_valid_sheet_name
from mitosheet.state import State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils.utils import get_param


def get_export_to_csv_sheet_index_to_file_name(file_name: str, sheet_indexes: List[int]) -> Dict[int, str]:
    """
    Note that we add the index to the file name if there are multiple sheets being exported,
    and we do not add the df name. 

    This is because if you are parameterizing this code chunk, and then you change the imported
    data, the file name will change, which will break the parameterization.

    So we need to make the file name independent of the df name, so that changing the df name does
    not result in a broken parameterization.
    """
    if len(sheet_indexes) == 1:
        return {sheet_indexes[0]: f'{file_name}'}
    else:
        file_name_without_extension, extension = os.path.splitext(file_name)
        return {sheet_index: f'{file_name_without_extension}_{index}{extension}' for index, sheet_index in enumerate(sheet_indexes)}

def get_export_to_excel_sheet_index_to_sheet_name(state: State, file_name: str, sheet_indexes: List[int]) -> Dict[int, str]:
    return {sheet_index: get_df_name_as_valid_sheet_name(state.df_names[sheet_index]) for sheet_index in sheet_indexes}


def get_final_file_name(file_name: str, _type: str) -> str:
    _, extension = os.path.splitext(file_name)
    if extension == '':
        if _type == 'csv':
            return file_name + '.csv'
        elif _type == 'excel':
            return file_name + '.xlsx'
    
    return file_name


def upgrade_export_to_file_1_to_2(step: Dict[str, Any], later_steps: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    '''
    Adds a new param for exporting the formatting to excel which defaults to false. 
    '''
    params = step['params']
    params['export_formatting'] = False
    return [{
        "step_version": 2, 
        "step_type": "export_to_file", 
        "params": params
    }] + later_steps

class ExportToFileStepPerformer(StepPerformer):
    """
    Allows you to export to file.
    """

    @classmethod
    def step_version(cls) -> int:
        return 2

    @classmethod
    def step_type(cls) -> str:
        return 'export_to_file'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:

        _type: str = get_param(params, 'type')
        sheet_indexes: List[int] = get_param(params, 'sheet_indexes')
        _file_name: str = get_param(params, 'file_name')

        file_name = get_final_file_name(_file_name, _type) # Ensure that the file name has the correct extension
        
        # We make a new state to modify it
        post_state = prev_state.copy()

        pandas_start_time = perf_counter()

        if _type == 'csv':
            sheet_index_to_export_location = get_export_to_csv_sheet_index_to_file_name(file_name, sheet_indexes)
        elif _type == 'excel':
            sheet_index_to_export_location = get_export_to_excel_sheet_index_to_sheet_name(post_state, file_name, sheet_indexes)
        else:
            raise ValueError(f"Invalid file type: {_type}")

        pandas_processing_time = perf_counter() - pandas_start_time

        return post_state, {
            'pandas_processing_time': pandas_processing_time,
            'sheet_index_to_export_location': sheet_index_to_export_location,
            'sheet_index_to_df_name': {i: n for i, n in enumerate(post_state.df_names)},
            'file_name': file_name,
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            ExportToFileCodeChunk(
                prev_state, 
                get_param(params, 'type'),
                get_param(execution_data if execution_data is not None else {}, 'file_name'),
                get_param(execution_data if execution_data is not None else {}, 'sheet_index_to_export_location'),
                get_param(execution_data if execution_data is not None else {}, 'sheet_index_to_df_name'),
                get_param(params, 'export_formatting')
            )
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return set()
    


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from datetime import datetime, timedelta
from typing import Optional, Union

import pandas as pd

def cast_to_timedelta(unknown: Union[str, int, float, bool, datetime, timedelta]) -> Optional[timedelta]:
    if isinstance(unknown, str):
        return pd.to_timedelta(unknown)
    elif isinstance(unknown, int):
        return pd.to_timedelta(unknown)
    elif isinstance(unknown, float):
        return pd.to_timedelta(unknown)
    elif isinstance(unknown, timedelta):
        return unknown

    return None


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import string
from typing import Tuple

from mitosheet.errors import make_invalid_range_error


def get_excel_range_from_column_index(col_index: int) -> str:
    """
    Number to Excel-style column name, e.g., 1 = A:A, 26 = Z:Z, 27 = AA:AA, 703 = AAA:AAA.
    """
    return get_column_from_column_index(col_index) + ":" + get_column_from_column_index(col_index)

def get_column_from_column_index(col_index: int) -> str:
    """
    Number to Excel-style column name, e.g., 1 = A, 26 = Z, 27 = AA, 703 = AAA.
    """
    # Add 1 because Mito 0 indexes columns
    col_index = col_index + 1
    name = ''
    while col_index > 0:
        col_index, r = divmod (col_index - 1, 26)
        name = chr(r + ord('A')) + name
    return name

def get_col_and_row_indexes_from_range(range: str) -> Tuple[Tuple[int, int], Tuple[int, int]]:
    try:
        start_cell_address, end_cell_address = range.split(':')
        (start_col_idx, start_row_idx) = get_row_and_col_index_from_cell_address(start_cell_address)
        (end_col_idx, end_row_idx) = get_row_and_col_index_from_cell_address(end_cell_address)
    except:
        raise make_invalid_range_error(range, False)

    return ((min(start_col_idx, end_col_idx), min(start_row_idx, end_row_idx)), (max(start_col_idx, end_col_idx), max(start_row_idx, end_row_idx)))

def get_row_and_col_index_from_cell_address(cell_address: str) -> Tuple[int, int]:
    
    letters = ''
    numbers = ''
    for char in cell_address:
        if char.isalpha():
            letters += char
        else:
            numbers += char

    return (get_index_from_column(letters), int(numbers) - 1)

def get_index_from_column(col: str) -> int:
    num = 0
    for c in col:
        if c in string.ascii_letters:
            num = num * 26 + (ord(c.upper()) - ord('A')) + 1
    return num - 1

def get_df_name_as_valid_sheet_name(df_name: str) -> str:
    return df_name[:31] # not more than 32 chars



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import Any, List, Tuple

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.state import State
from mitosheet.transpiler.transpile_utils import \
    get_column_header_as_transpiled_code


class PromoteRowToHeaderCodeChunk(CodeChunk):

    def __init__(self, prev_state: State, sheet_index: int, index: Any):
        super().__init__(prev_state)
        self.sheet_index = sheet_index
        self.index = index

        self.df_name = self.prev_state.df_names[self.sheet_index]


    def get_display_name(self) -> str:
        return 'Promote Row To Header'
    
    def get_description_comment(self) -> str:

        return f"Promoted row {self.index} to header in {self.df_name}"
        
    def get_code(self) -> Tuple[List[str], List[str]]:
        transpiled_index = get_column_header_as_transpiled_code(self.index)

        code = [f"{self.df_name}.columns = {self.df_name}.loc[{transpiled_index}]"]

        code.append(f"{self.df_name}.columns = deduplicate_column_headers({self.df_name}.columns.tolist())")

        code.append(f"{self.df_name}.drop(labels=[{transpiled_index}], inplace=True)")
        

        return code, []
    
    def get_edited_sheet_indexes(self) -> List[int]:
        return [self.sheet_index]
    


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from datetime import datetime, timedelta
from typing import Any

import numpy as np
import pandas as pd

from mitosheet.errors import MitoError
from mitosheet.is_type_utils import (is_bool_dtype, is_datetime_dtype,
                                     is_float_dtype, is_int_dtype,
                                     is_string_dtype)
from mitosheet.public.v3.errors import handle_sheet_function_errors
from mitosheet.public.v3.sheet_functions.utils import \
    get_series_from_primitive_or_series
from mitosheet.public.v3.types.decorators import cast_values_in_arg_to_type
from mitosheet.public.v3.types.sheet_function_types import (
    AnyPrimitiveOrSeriesInputType, BoolRestrictedInputType,
    IntRestrictedInputType)


@handle_sheet_function_errors
def FILLNAN(series: pd.Series, replacement: AnyPrimitiveOrSeriesInputType) -> pd.Series:
    """
    {
        "function": "FILLNAN",
        "description": "Replaces the NaN values in the series with the replacement value.",
        "search_terms": ["fillnan", "nan", "fill nan", "missing values", "null", "null value", "fill null"],
        "examples": [
            "FILLNAN(A, 10)",
            "FILLNAN(A, 'replacement')"
        ],
        "syntax": "FILLNAN(series, replacement)",
        "syntax_elements": [{
                "element": "series",
                "description": "The series to replace the NaN values in."
            },
            {
                "element": "replacement",
                "description": "A string, number, or date to replace the NaNs with."
            }
        ]
    }
    """
    return series.fillna(replacement)


@handle_sheet_function_errors
def TYPE(series: pd.Series) -> pd.Series:
    """
    {
        "function": "TYPE",
        "description": "Returns the type of each element of the passed series. Return values are 'number', 'str', 'bool', 'datetime', 'object', or 'NaN'.",
        "search_terms": ["type", "dtype"],
        "examples": [
            "TYPE(Nums_and_Strings)",
            "IF(TYPE(Account_Numbers) != 'NaN', Account_Numbers, 0)"
        ],
        "syntax": "TYPE(series)",
        "syntax_elements": [{
                "element": "series",
                "description": "The series to get the type of each element of."
            }
        ]
    }
    """

    def get_element_type(element):
        # Start with bool!
        if isinstance(element, bool):
            return 'bool'
        elif isinstance(element, int):
            return 'number'
        elif isinstance(element, float):
            if np.isnan(element):
                return 'NaN'
            return 'number'
        elif element is None or element is pd.NaT:
            return 'NaN'
        elif isinstance(element, str):
            return 'string'
        elif isinstance(element, datetime) or isinstance(element, pd.Timestamp):
            return 'datetime'
        elif isinstance(element, timedelta) or isinstance(element, pd.Timedelta):
            return 'timedelta'
        return 'object'

    return series.apply(get_element_type).astype('str')



@handle_sheet_function_errors
def GETPREVIOUSVALUE(series: pd.Series, condition: BoolRestrictedInputType) -> pd.Series:
    """
    {
        "function": "GETPREVIOUSVALUE",
        "description": "Returns the value from series that meets the condition.",
        "search_terms": ["ffill"],
        "examples": [
            "GETPREVIOUSVALUE(Max_Balances, Max_Balances > 0)"
        ],
        "syntax": "GETPREVIOUSVALUE(series, condition)",
        "syntax_elements": [{
                "element": "series",
                "description": "The series to get the previous value from."
            }, {
                "element": "condition",
                "description": "When condition is True, a new previous value is set, and carried forward until the condition is True again."
            }
        ]
    }
    """
    # Default to a different last occurence depending on the type
    column_dtype = str(series.dtype)
    last_occurrence: Any = -1
    if is_int_dtype(column_dtype) or is_float_dtype(column_dtype):
        last_occurrence = -1
    elif is_string_dtype(column_dtype):
        last_occurrence = ''
    elif is_bool_dtype(column_dtype):
        last_occurrence = False
    elif is_datetime_dtype(column_dtype):
        last_occurrence = pd.NaT

    condition = get_series_from_primitive_or_series(condition, series.index)

    result = []
    for index, value in condition.items():
        if value:
            last_occurrence = series[index]
        result.append(last_occurrence)

    return pd.Series(result, index=series.index)

@handle_sheet_function_errors
def GETNEXTVALUE(series: pd.Series, condition: BoolRestrictedInputType) -> pd.Series:
    """
    {
        "function": "GETNEXTVALUE",
        "description": "Returns the next value from series that meets the condition.",
        "search_terms": ["ffill"],
        "examples": [
            "GETNEXTVALUE(Max_Balances, Max_Balances > 0)"
        ],
        "syntax": "GETNEXTVALUE(series, condition)",
        "syntax_elements": [{
                "element": "series",
                "description": "The series to get the next value from."
            }, {
                "element": "condition",
                "description": "When condition is True, a new previous value is set, and carried backwards until the condition is True again."
            }
        ]
    }
    """
    reversed_series = series[::-1]
    condition = get_series_from_primitive_or_series(condition, series.index)
    reversed_condition = condition[::-1]

    return GETPREVIOUSVALUE(reversed_series, reversed_condition)[::-1]

@cast_values_in_arg_to_type('index', 'int')
def VLOOKUP(lookup_value: AnyPrimitiveOrSeriesInputType, where: pd.DataFrame, index: IntRestrictedInputType) -> pd.Series:
    """
    {
        "function": "VLOOKUP",
        "description": "Looks up a value in a range and returns the value in the same row from a column you specify.",
        "search_terms": ["vlookup", "merge", "join", "search", "lookup"],
        "category": "REFERENCE",
        "examples": [
            "VLOOKUP(Names0, Ids:Ages, 1)",
            "VLOOKUP('John Smith', Names:Ages, 2)",
            "VLOOKUP(Names0, Ids:Ages, Column Indexes0)"
        ],
        "syntax": "VLOOKUP(lookup_value, where, index)",
        "syntax_elements": [{
                "element": "lookup_value",
                "description": "The value to look up."
            }, {
                "element": "where",
                "description": "The range to look up in."
            }, {
                "element": "index",
                "description": "The column index to return."
            }
        ]
    }
    """
    where_first_column_case_insensitive = where.iloc[:,0].copy()
    where_first_column_case_insensitive.name = str(where_first_column_case_insensitive.name) + 'MITO_CASE_INSENSITIVE'

    # If the lookup value and index are both a primitive, we don't need to merge. 
    if not isinstance(lookup_value, pd.Series) and isinstance(index, int):
        if type(lookup_value) != type(where.iloc[0,0]):
            raise MitoError(
                'invalid_args_error',
                'VLOOKUP',
                f'VLOOKUP requires the lookup value and the first column of the where range to be the same type. The lookup value is of type {type(lookup_value)} and the first column of the where range is of type {type(where.iloc[0,0])}.'
            )

        # If the lookup value and the first column are strings, make them lowecase for case-insensitive matching
        if isinstance(lookup_value, str) and isinstance(where.iloc[0,0], str):
            lookup_value = lookup_value.lower()
            where_first_column_case_insensitive = where_first_column_case_insensitive.str.lower()

        matching_row = where.loc[where_first_column_case_insensitive == lookup_value]
        if matching_row.empty:
            return None
        else:
            return matching_row.iloc[0, index-1]

    value = get_series_from_primitive_or_series(lookup_value, where.index)

    value.name = 'lookup_value'
    indices_to_return_from_range = get_series_from_primitive_or_series(index, value.index)

    # If the lookup value and the first column of the where range are different types, we raise an error
    if value.dtype != where_first_column_case_insensitive.dtype:
        raise MitoError(
            'invalid_args_error',
            'VLOOKUP',
            f'VLOOKUP requires the lookup value and the first column of the where range to be the same type. The lookup value is of type {value.dtype} and the first column of the where range is of type {where_first_column_case_insensitive.dtype}.'
        )

    # If the series is a string, convert it to lowercase because Excel's vlookup is case insensitive
    if is_string_dtype(str(value.dtype)):
        value = value.str.lower()
        where_first_column_case_insensitive = where_first_column_case_insensitive.str.lower()

    # Add where_first_column_case_insensitive to the front of the dataframe so we can use the case insensitive merge 
    # without effecting the return values
    where = pd.concat([where_first_column_case_insensitive, where], axis=1)
    indices_to_return_from_range = indices_to_return_from_range + 1

    where_deduplicated = where.drop_duplicates(subset=where_first_column_case_insensitive.name)
    
    # Update first column to use the deduplicated version
    where_first_column = where_deduplicated.iloc[:,0]
    
    # Then merge on the column we're looking up from, and the df we're looking up in.
    merged = pd.merge(value, where_deduplicated, left_on='lookup_value', right_on=where_first_column, how='left')

    # Change the indexes back to the indexes of the lookup value so the results 
    # can be added back to the calling dataframe.
    merged.index = value.index
    
    def get_value_at_index_in_row(row):
        try:
            return row.iloc[indices_to_return_from_range[row.name]]
        # Because we can't control what the user puts in the index, we need to catch any errors
        except Exception:
            return None
    return merged.apply(get_value_at_index_in_row, axis=1)

# TODO: we should see if we can list these automatically!
MISC_FUNCTIONS = {
    'FILLNAN': FILLNAN,
    'GETPREVIOUSVALUE': GETPREVIOUSVALUE,
    'GETNEXTVALUE': GETNEXTVALUE,
    'TYPE': TYPE,
    'VLOOKUP': VLOOKUP
}


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from datetime import datetime, timedelta
from typing import Optional, Union

from mitosheet.public.v3.errors import make_invalid_param_type_conversion_error
from mitosheet.public.v3.types.float import cast_string_to_float


def cast_str_to_int(s: str) -> Optional[int]:
    f = cast_string_to_float(s)
    if f is None:
        raise make_invalid_param_type_conversion_error(s, 'int')
    else:
        return int(f)

def cast_to_int(unknown: Union[str, int, float, bool, datetime, timedelta]) -> Optional[int]:
    if isinstance(unknown, bool):
        return int(unknown)
    elif isinstance(unknown, str):
        return cast_str_to_int(unknown)
    elif isinstance(unknown, int):
        return unknown
    elif isinstance(unknown, float):
        try:
            return int(unknown)
        except:
            raise make_invalid_param_type_conversion_error(unknown, 'int')

    return None


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import json
import logging
import time
from typing import Any, Dict, List, Optional
import threading

import requests

"""
Error Severity Codes are used to help enterprise admins triage errors. 
When managing the Python environment of 1000's of users, error severity codes 
are useful for understanding the impact of errors and prioritizing fixes.

The error severity codes use the following schema:

Likely just a warning		        0
Likely user error		            10
Likely inconsequential error		11
Likely Mito Bug		                20
Unable to import data		        21
Unable to replay analysis		    22	
Mito Crashed		                50	
Misc.		                        -1
"""

error_severity_codes = {
    'ai_transformation_edit_failed': '10',
    'append_user_field_update_failed': '11',
    'args_update_failed': '20',
    'args_update_remains_failed': '20',
    'change_column_dtype_edit_failed': '10',
    'check_dependencies_failed': '-1',
    'dataframe_delete_edit_failed': '11',
    'dataframe_import_edit_failed': '21',
    'dataframe_rename_edit_failed': '11',
    'delete_column_edit_failed': '11',
    'delete_row_edit_failed': '11',
    'excel_import_edit_failed': '21',
    'filter_column_edit_failed': '10',
    'frontend_render_failed': '50',
    'get_column_summary_graph_failed': '11',
    'get_dataframe_as_excel_failed': '21',
    'get_defined_df_names_failed': '21',
    'get_excel_file_metadata_failed': '21',
    'get_unique_value_counts_failed': '20',
    'graph_edit_failed': '10',
    'graph_rename_edit_failed': '11',
    'install_failed': '-1',
    'install_mitosheet_failed': '-1',
    'melt_edit_failed': '11',
    'merge_edi_failed':' 0',
    'mitosheet_sheet_call_failed': '50',
    'mitosheet_sheet_call_location_failed': '10',
    'one_hot_encoding_edit_failed': '11',
    'overwrite_analysis_to_replay_to_mitosheet_call_failed': '22',
    'pivot_edit_failed': '20',
    'plotly_define_failed': '20',
    'rename_column_edit_failed': '10',
    'render_count_update_failed': '11',
    'replace_edit_failed': '10',
    'replay_analysis_update_failed': '22',
    'replayed_nonexistant_analysis_failed': '22',
    'set_column_formula_edit_failed': '10',
    'simple_import_edit_failed': '21',
    'sort_edit_failed': '20',
    'write_analysis_to_replay_to_mitosheet_call_failed': '20'
}


def preprocess_log_for_upload(log_event: str, log_params: Dict[str, Any]) -> Optional[Dict[str, Any]]:
    """
    Convert logs into the correct format and remove any log events that are not part of the whitelisted schema.

	{
	    'timestamp': '2023-10-25T15:30:00Z',
	    'event': 'set_column_formula',
	    'params': {
            'sheet_index': 1,
            'column_id': 'column id',
            'new_formula': '=10 + 11'
	    },
        'version_python': '3.9',
        'version_pandas': '2.0',
        'version_mitosheet': '0.1.522',
        'error_severity_code': '10'
	}
    """
    
    whitelisted_log_events = [
        'edit_event',
        'error', 
        'mitosheet_rendered',
        'frontend_render_failed',
    ]

    whitelisted_log_params = [
        'version_python',
        'version_pandas',
        'version_mito',
        'error_traceback',
        'error_traceback_last_line',
    ]

    # Remove non-whitelisted events
    if log_event not in whitelisted_log_events:
        return None

    # Remove any log params that are not part of whitelisted params or start with "params", ie: params_sheet_index
    filtered_log_params = {k: v for k, v in log_params.items() if k in whitelisted_log_params or k.startswith('params_')}

    # Add the gmt timestamp formatted as 2023-10-25T15:30:00Z
    filtered_log_params['timestamp_gmt'] = time.strftime('%Y-%m-%dT%H:%M:%SZ', time.gmtime())

    # Create a useful top-level event field. Use the params_log_event if it exists (edit_events)
    # Otherwise, use the log_event.
    if 'params_log_event' in filtered_log_params:
        filtered_log_params['event'] = filtered_log_params['params_log_event']
        del filtered_log_params['params_log_event']
    else:
        filtered_log_params['event'] = log_event

    # Add the severity code to the log and default to -1 if not found
    if log_event == 'error':
        filtered_log_params['error_severity_code'] = error_severity_codes.get(filtered_log_params['params_failed_log_event'], '-1')
    if log_event == 'frontend_render_failed':
        filtered_log_params['error_severity_code'] = error_severity_codes.get(log_event, '-1')

    return filtered_log_params

# Set to True if you want to run the log uploader in a threaded environment
# NOTE: because the log uploader only works in JupyterLab, we do not need
# to ever handle the case where LOG_UPLOADER_IS_THREADED is False for now
# but we may need it in the future if we expand to other environments
LOG_UPLOADER_IS_THREADED = True

class MitoLogUploader:
    """
    The MitoLogUploader is responsible for uploading logs to a 
    custom analytics url set in the MITO_CONFIG.

    It only uploads the most useful logs and log params in order 
    to make the logs maximally useful and minimally confusing. 
     
    It also uploads logs in batches every log_interval seconds so that 
    enterprises collecting logs from thousands of users do not 
    break their logging server. 

    Note that we use an exponential backoff strategy to prevent overloading
    the logging server in the case of a temporary failure.
    """
    def __init__(self, log_url: str, log_interval: Optional[int]):
        self.log_url = log_url
        self.base_log_interval = log_interval if log_interval is not None else 10
        self.current_log_interval = self.base_log_interval
        self.last_upload_time = time.time()
        self.unprocessed_logs: List[Dict[str, Any]] = []
        self.lock = threading.Lock()  # Lock for thread-safe operations
        
        if LOG_UPLOADER_IS_THREADED:
            self.upload_thread = threading.Thread(target=self.__run_periodic_upload, daemon=True)
            self.upload_thread.start()

    def log(self, log_event: str, log_params: Dict[str, Any]) -> None:
        """
        Converts log into the correct format, adds it to the queue of logs to be uploaded,
        and checks if it is time to upload the logs.
        """
        with self.lock:  # Ensure thread-safe access to shared resources
            filtered_log_params = preprocess_log_for_upload(log_event, log_params)
            if filtered_log_params is not None:
                self.unprocessed_logs.append(filtered_log_params)

        # We only upload logs if we are not running in a threaded environment
        # In a threaded environment, we upload logs every log_interval seconds
        # in the run_periodic_upload function
        if not LOG_UPLOADER_IS_THREADED:
            current_time = time.time()
            if self.last_upload_time + self.current_log_interval < current_time and len(self.unprocessed_logs) > 0:
                self.__try_upload_logs()

    def __try_upload_logs(self) -> None:
        """
        Uploads the unprocessed logs to the log_url and clears the unprocessed logs.

        If the logs are not being uploaded successfully, then double the log interval
        This is an exponential backoff strategy to prevent overloading the logging server.
        """
        with self.lock:  # Protect access to shared resources
            self.last_upload_time = time.time()
            
            if len(self.unprocessed_logs) == 0:
                return  # No logs to upload

            log_payload = json.dumps(self.unprocessed_logs)

            try:
                result = requests.post(
                    self.log_url,
                    data=log_payload,
                    headers={'Content-Type': 'application/json'}
                )
                # Throw an error if the request was not successful
                result.raise_for_status()

                self.unprocessed_logs = []
                
                # Per the exponential backoff strategy, if the logs are being uploaded successfully,
                # then reset the log interval to the base interval
                self.current_log_interval = self.base_log_interval

            except Exception as e:
                print(f"Log upload failed with error:", e, flush=True)

                # If the logs are not being uploaded successfully, then double the log interval
                # This is an exponential backoff strategy to prevent overloading the logging server
                # in the case of a temporary failure
                self.current_log_interval *= 2

    def __run_periodic_upload(self):
        while LOG_UPLOADER_IS_THREADED:
            time.sleep(self.current_log_interval)
            self.__try_upload_logs()


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

"""
Utilities for figuring out where Mito is being run.
"""

import os

from mitosheet.utils import run_command


def is_in_google_colab() -> bool:
    # From: https://discourse.jupyter.org/t/find-out-if-my-code-runs-inside-a-notebook-or-jupyter-lab/6935/19
    try:
        from IPython.core.getipython import get_ipython
        return 'google.colab' in str(get_ipython()) # type: ignore
    except:
        return False

def is_in_vs_code() -> bool:
    # From: https://github.com/microsoft/vscode-jupyter/issues/3364
    return 'VSCODE_PID' in os.environ

def is_jupyter_lab_running() -> bool:
    """
    Format of command output: 

    Currently running servers:
    http://localhost:8888/?token=wenflwenflqnwdlkqmwldkm :: Path/to/launch
    """
    try:
        return '::' in run_command(['jupyter', 'lab', 'list'])[0]
    except:
        return False

def is_jupyter_notebook_running() -> bool:
    """
    Format of command output: 

    Currently running servers:
    http://localhost:8888/?token=wenflwenflqnwdlkqmwldkm :: Path/to/launch
    """
    try:
        return '::' in run_command(['jupyter', 'notebook', 'list'])[0]
    except:
        return False

def is_notebook() -> bool:
    """
    Returns true if this code is run in some sort of notebook
    Taken from: https://stackoverflow.com/questions/15411967/how-can-i-check-if-code-is-executed-in-the-ipython-notebook
    """
    try:
        from IPython.core.getipython import get_ipython
    except:
        return False
    try:
        shell = get_ipython().__class__.__name__ # type: ignore
        if shell == 'ZMQInteractiveShell':
            return True   # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False      # Probably standard Python interpreter

def get_location() -> str:
    notebook = is_notebook()
    lab_running = is_jupyter_lab_running()

    if is_in_google_colab():
        return 'location_google_colab'
    elif is_in_vs_code():
        return 'location_vs_code'
    elif is_streamlit():
        return 'location_streamlit'
    elif is_dash():
        return 'location_dash'
    elif is_jupyterlite():
        return 'location_jupyterlite'
    elif notebook and lab_running:
        # NOTE: Since Notebook 7 runs on jlab, we can't easily 
        # tell if the user is viewing the notebook interface or the lab,
        # so we just say that they are in jupyter.
        return 'location_jupyter'
    elif notebook:
        # NOTE: in this case, they are not using jlab so we don't 
        # know what notebook they are in. So we just say unknown
        return 'location_unknown_notebook'
    else:
        return 'location_unknown'


# https://stackoverflow.com/questions/43878953/how-does-one-detect-if-one-is-running-within-a-docker-container-within-python
def _is_docker() -> bool:
    path = '/proc/self/cgroup'
    return (
        os.path.exists('/.dockerenv') or
        os.path.isfile(path) and any('docker' in line for line in open(path))
    )
# We cache this so we don't have to take a performance impact
IS_DOCKER = _is_docker()
def is_docker() -> bool:
    return IS_DOCKER



def is_jupyterlite() -> bool:
    """
    We do our best to guess if you are in jupyterlite, if:
    1. You have pyodide installed
    2. You cannot start a thread
    """
    try:
        import pyodide
        
        try:
            import threading
            threading.Thread(target=lambda x: x).start()
        except RuntimeError:
            return True

    except ImportError:
        pass

    return False

def is_streamlit() -> bool:
    try:
        # TODO: have to handle versions of streamlit that don't have this
        # or it's in different places: https://discuss.streamlit.io/t/how-to-check-if-code-is-run-inside-streamlit-and-not-e-g-ipython/23439/8
        from streamlit.runtime.scriptrunner import get_script_run_ctx
        return get_script_run_ctx() is not None
    except ModuleNotFoundError:
        return False
    
_IS_DASH = False
    
def is_dash() -> bool:
    global _IS_DASH
    # If we are ever in dash, we always are, so we can avoid work. Notably, because the dash app starts after the Mito Backend is created
    # we do have to check multiple times. 
    if _IS_DASH:
        return True
    
    try:
        
        import dash
        from flask import current_app

        # If any of these routes are defined, then we are in a dash app
        dash_routes = [
            '/_dash-dependencies',
            '/_dash-component-suites',
            '/_dash-layout'
        ]

        for rule in current_app.url_map.iter_rules():
            for dash_route in dash_routes:
                if dash_route in str(rule):
                    return True

    except:
        pass
    return False


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from mitosheet.mito_dash.v1 import Spreadsheet, activate_mito
from dash import Dash, html
import pandas as pd

app = Dash(__name__)
activate_mito(app)

app.layout = html.Div([
    html.H1('Empty Mito Spreadsheet'),
    Spreadsheet(id={'type': 'spreadsheet', 'id': 'sheet'}, import_folder='datasets')
])

if __name__ == '__main__':
    app.run_server(debug=True)


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from time import perf_counter
from typing import Any, Dict, List, Optional, Set, Tuple

import pandas as pd

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.empty_code_chunk import EmptyCodeChunk
from mitosheet.state import State
from mitosheet.step_performers.graph_steps.graph_utils import (
    get_column_header_from_optional_column_id_graph_param,
    get_graph_index_by_graph_id,
    get_html_and_script_from_figure, get_new_graph_tab_name)
from mitosheet.step_performers.graph_steps.plotly_express_graphs import (
    get_plotly_express_graph, get_plotly_express_graph_code)
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils.utils import get_param
from mitosheet.types import GraphID


class GraphStepPerformer(StepPerformer):
    """
    Creates a graph of the passed parameters and update the graph_data_array

    {
        graph_id: GraphID
        graph_preprocessing: {
            safety_filter_turned_on_by_user: boolean
        },
        graph_creation: {
            graph_type: GraphType,
            sheet_index: int
            x_axis_column_ids: ColumnID[],
            y_axis_column_ids: ColumnID[],
            color: columnID: columnID | None,
            facet_column: columnID | None,
            facet_row: columnID | None,
        },
        graph_styling: {
            title: {
                title: string | None
                visible: boolean
            },
            xaxis: {
                title: string | None,
                visible: boolean,
                type: string | None,
                rangeslider: {
                    visible: boolean
                }
            },
            yaxis: {
                title: string | None,
                visible: boolean,
                type: string | None
            },
            showlegend: boolean
        },
        graph_rendering: {
            height: int representing the div width
            width: int representing the div width
        }
    }
    """

    @classmethod
    def step_version(cls) -> int:
        return 4

    @classmethod
    def step_type(cls) -> str:
        return "graph"

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        """
        Returns the new post state with the updated graph_data_array
        """

        graph_id: GraphID = get_param(params, 'graph_id')
        graph_preprocessing: Dict[str, Any] = get_param(params, 'graph_preprocessing')
        graph_creation: Dict[str, Any] = get_param(params, 'graph_creation')
        graph_styling: Dict[str, Any] = get_param(params, 'graph_styling')
        graph_rendering: Dict[str, Any] = get_param(params, 'graph_rendering')

        include_plotlyjs: bool = get_param(params, 'include_plotlyjs')
        
        # We make a new state to modify it
        post_state = prev_state.copy()

        # Extract variables from graph parameters
        graph_type = graph_creation["graph_type"]
        sheet_index = graph_creation["sheet_index"]
        safety_filter_turned_on_by_user = graph_preprocessing["safety_filter_turned_on_by_user"]
        height: str = graph_rendering["height"]
        width: str = graph_rendering["width"]

        # Get the x axis params, if they were provided
        x_axis_column_ids = graph_creation["x_axis_column_ids"] if graph_creation["x_axis_column_ids"] is not None else []
        x_axis_column_headers = prev_state.column_ids.get_column_headers_by_ids(sheet_index, x_axis_column_ids)

        # Get the y axis params, if they were provided
        y_axis_column_ids = graph_creation["y_axis_column_ids"] if graph_creation["y_axis_column_ids"] is not None else []
        y_axis_column_headers = prev_state.column_ids.get_column_headers_by_ids(sheet_index, y_axis_column_ids)

        # Validate optional parameters that are available for all graph types
        color_column_header = get_column_header_from_optional_column_id_graph_param(prev_state, graph_creation, 'color')
        facet_col_column_header = get_column_header_from_optional_column_id_graph_param(prev_state, graph_creation, 'facet_col_column_id')
        facet_row_column_header = get_column_header_from_optional_column_id_graph_param(prev_state, graph_creation, 'facet_row_column_id')
        
        facet_col_wrap = graph_creation.get('facet_col_wrap', None)
        facet_col_spacing = graph_creation.get('facet_col_spacing', None)
        facet_row_spacing = graph_creation.get('facet_row_spacing', None)

        # Validate parameters that are only available for some graph types
        # Note: We trust the parameters from the frontend, so we don't make sure the params fit the graph type here
        points = graph_creation.get('points', None)
        line_shape = graph_creation.get('line_shape', None)
        histnorm = graph_creation.get('histnorm', None)
        histfunc = graph_creation.get('histfunc', None)
        nbins = graph_creation.get('nbins', None)

        # Create a copy of the dataframe, just for safety.
        df: pd.DataFrame = prev_state.dfs[sheet_index].copy()
        df_name: str = prev_state.df_names[sheet_index]

        # If the graph tab already exists, use its name. Otherwise, create a new graph tab name.
        graph_index = get_graph_index_by_graph_id(post_state.graph_data_array, graph_id)
        graph_tab_name: str = post_state.graph_data_array[graph_index]["graph_tab_name"] \
            if graph_index != -1 \
            else get_new_graph_tab_name(post_state.graph_data_array)
        if graph_index == -1:
            post_state.graph_data_array.append({
                "graph_id": graph_id,
                "graph_tab_name": graph_tab_name
            })
            graph_index = len(post_state.graph_data_array) - 1

        if len(x_axis_column_ids) == 0 and len(y_axis_column_ids) == 0:
            # If no data is passed to the graph, then we don't create a graph and omit the graphOutput
            post_state.graph_data_array[graph_index] = {
                "graph_id": graph_id,
                "graph_tab_name": graph_tab_name
            }
            pandas_processing_time = 0.0 # no processing time
        else: 
            pandas_start_time = perf_counter()
            fig = get_plotly_express_graph(
                graph_type,
                df,
                safety_filter_turned_on_by_user,
                x_axis_column_headers,
                y_axis_column_headers,
                color_column_header,
                facet_col_column_header,
                facet_row_column_header,
                facet_col_wrap,
                facet_col_spacing,
                facet_row_spacing,
                points,
                line_shape,
                histnorm,
                histfunc,
                nbins,
                graph_styling
            )
            pandas_processing_time = perf_counter() - pandas_start_time

            # Get rid of some of the default white space
            fig.update_layout(
                margin=dict(
                    l=0,
                    r=0,
                    t=30,
                    b=35, # This gives enough space so that the x axis label is not cutoff
                )
            )

            html_and_script = get_html_and_script_from_figure(fig, height, width, include_plotlyjs)

            graph_generation_code = get_plotly_express_graph_code(
                graph_type,
                df,
                safety_filter_turned_on_by_user,
                x_axis_column_headers,
                y_axis_column_headers,
                color_column_header,
                facet_col_column_header,
                facet_row_column_header,
                facet_col_wrap,
                facet_col_spacing,
                facet_row_spacing,
                points,
                line_shape,
                histnorm,
                histfunc,
                nbins,
                graph_styling,
                df_name,
            )

            post_state.graph_data_array[graph_index] = {
                "graph_id": graph_id,
                "graph_output": {
                    "graphGeneratedCode": graph_generation_code,
                    "graphHTML": html_and_script["html"],
                    "graphScript": html_and_script["script"],
                },
                "graph_tab_name": graph_tab_name
            }

        return post_state, {
            'pandas_processing_time': pandas_processing_time # No time spent on pandas, only metadata changes
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        # Graph steps don't add any generated code to the analysis script. 
        return [
            EmptyCodeChunk(
                prev_state, 
                'Edited graph',
                'Edited a graph',
            )
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {-1}



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the The Mito Enterprise license.

from ast import List
import datetime
import pprint
from typing import Any, Callable, Dict, Optional, Union
import os
from mitosheet.enterprise.license_key import decode_license_to_date
from mitosheet.telemetry.telemetry_utils import log
from mitosheet.types import CodeSnippetEnvVars
from mitosheet.user.utils import is_enterprise

# Note: Do not change these keys, we need them for looking up 
# the environment variables from previous mito_config_versions.
MITO_CONFIG_VERSION = 'MITO_CONFIG_VERSION'
MITO_CONFIG_SUPPORT_EMAIL = 'MITO_CONFIG_SUPPORT_EMAIL'
MITO_CONFIG_CODE_SNIPPETS_SUPPORT_EMAIL = 'MITO_CONFIG_CODE_SNIPPETS_SUPPORT_EMAIL' 
MITO_CONFIG_CODE_SNIPPETS_VERSION  = 'MITO_CONFIG_CODE_SNIPPETS_VERSION' 
MITO_CONFIG_CODE_SNIPPETS_URL = 'MITO_CONFIG_CODE_SNIPPETS_URL'
MITO_CONFIG_DISABLE_TOURS = 'MITO_CONFIG_DISABLE_TOURS'
MITO_CONFIG_FEATURE_ENABLE_SNOWFLAKE_IMPORT = 'MITO_CONFIG_FEATURE_ENABLE_SNOWFLAKE_IMPORT'
MITO_CONFIG_FEATURE_DISPLAY_SNOWFLAKE_IMPORT = 'MITO_CONFIG_FEATURE_DISPLAY_SNOWFLAKE_IMPORT'
MITO_CONFIG_FEATURE_DISPLAY_AI_TRANSFORMATION = 'MITO_CONFIG_FEATURE_DISPLAY_AI_TRANSFORMATION'
MITO_CONFIG_FEATURE_DISPLAY_SCHEDULING = 'MITO_CONFIG_FEATURE_DISPLAY_SCHEDULING'
MITO_CONFIG_FEATURE_DISPLAY_CODE_OPTIONS = 'MITO_CONFIG_FEATURE_DISPLAY_CODE_OPTIONS'
MITO_CONFIG_FEATURE_TELEMETRY = 'MITO_CONFIG_FEATURE_TELEMETRY'
MITO_CONFIG_PRO = 'MITO_CONFIG_PRO'
MITO_CONFIG_ENTERPRISE = 'MITO_CONFIG_ENTERPRISE'
MITO_CONFIG_ENTERPRISE_TEMP_LICENSE = 'MITO_CONFIG_ENTERPRISE_TEMP_LICENSE'
MITO_CONFIG_LLM_URL = 'MITO_CONFIG_LLM_URL'
MITO_CONFIG_ANALYTICS_URL = 'MITO_CONFIG_ANALYTICS_URL'
MITO_CONFIG_CUSTOM_SHEET_FUNCTIONS_PATH = 'MITO_CONFIG_CUSTOM_SHEET_FUNCTIONS_PATH'
MITO_CONFIG_CUSTOM_IMPORTERS_PATH = 'MITO_CONFIG_CUSTOM_IMPORTERS_PATH'
MITO_CONFIG_LOG_SERVER_URL = 'MITO_CONFIG_LOG_SERVER_URL'
MITO_CONFIG_LOG_SERVER_BATCH_INTERVAL = 'MITO_CONFIG_LOG_SERVER_BATCH_INTERVAL'


# Note: The below keys can change since they are not set by the user.
MITO_CONFIG_CODE_SNIPPETS = 'MITO_CONFIG_CODE_SNIPPETS'

# The default values to use if the mec does not define them
DEFAULT_MITO_CONFIG_SUPPORT_EMAIL = 'founders@sagacollab.com'
DEFAULT_MITO_CONFIG_CODE_SNIPPETS_SUPPORT_EMAIL = 'founders@sagacollab.com'

# Since Mito needs to look up individual environment variables, we need to 
# know the names of the variables associated with each mito config version. 
# To do so we store them as a list here. 
MEC_VERSION_KEYS = {
    '1': [MITO_CONFIG_VERSION, MITO_CONFIG_SUPPORT_EMAIL],
    '2': [
        MITO_CONFIG_VERSION, 
        MITO_CONFIG_SUPPORT_EMAIL, 
        MITO_CONFIG_CODE_SNIPPETS_SUPPORT_EMAIL, 
        MITO_CONFIG_CODE_SNIPPETS_VERSION,
        MITO_CONFIG_CODE_SNIPPETS_URL, 
        MITO_CONFIG_DISABLE_TOURS,
        MITO_CONFIG_FEATURE_ENABLE_SNOWFLAKE_IMPORT,
        MITO_CONFIG_FEATURE_DISPLAY_SNOWFLAKE_IMPORT,
        MITO_CONFIG_FEATURE_DISPLAY_AI_TRANSFORMATION,
        MITO_CONFIG_FEATURE_DISPLAY_SCHEDULING,
        MITO_CONFIG_FEATURE_DISPLAY_CODE_OPTIONS,
        MITO_CONFIG_LLM_URL,
        MITO_CONFIG_ANALYTICS_URL,
        MITO_CONFIG_LOG_SERVER_URL,
        MITO_CONFIG_LOG_SERVER_BATCH_INTERVAL,
        MITO_CONFIG_FEATURE_TELEMETRY,
        MITO_CONFIG_PRO,
        MITO_CONFIG_ENTERPRISE,
        MITO_CONFIG_ENTERPRISE_TEMP_LICENSE,
        MITO_CONFIG_CUSTOM_SHEET_FUNCTIONS_PATH,
        MITO_CONFIG_CUSTOM_IMPORTERS_PATH,
    ]
}

def upgrade_mec_1_to_2(mec: Dict[str, Any]) -> Dict[str, Any]:
    """
    Converts mec of shape:
    {
        'MITO_CONFIG_VERSION': '1',
        'MITO_CONFIG_SUPPORT_EMAIL': 'support@mito.com'
    }
    
    into: 

    {
        'MITO_CONFIG_VERSION': '2',
        'MITO_CONFIG_SUPPORT_EMAIL': 'support@mito.com',
        'MITO_CONFIG_CODE_SNIPPETS_SUPPORT_EMAIL': None,
        'MITO_CONFIG_CODE_SNIPPETS_VERSION': None,
        'MITO_CONFIG_CODE_SNIPPETS_URL': None,
        'MITO_CONFIG_DISABLE_TOURS': None,
        'MITO_CONFIG_FEATURE_ENABLE_SNOWFLAKE_IMPORT': None,
        
    }
    """

    # Find the new keys in version 2 that are not in version 1
    new_keys = set(MEC_VERSION_KEYS['2']) - set(mec.keys())
    for key in new_keys:
        mec[key] = None

    mec[MITO_CONFIG_VERSION] = '2'
    return mec

"""
When updating the MEC_VERSION, add a function here to update the previous mec to the new version. For example, 
if mec_version='3', mec_upgrade_functions should look like:
{
   '1': upgrade_mec_1_to_2,
   '2': upgrade_mec_2_to_3
}
To keep things simple for now, these upgrade functions just make sure that all of the keys are defined so 
that the functions below can set the correct default values and format the mec properly.
"""

mec_upgrade_functions: Dict[str, Any] = {
    '1': upgrade_mec_1_to_2
}

def upgrade_mito_enterprise_configuration(mec: Optional[Dict[str, Any]]) -> Optional[Dict[str, Any]]:
    if mec is None:
        return None

    # So mypy tests recognize that mec is not None
    _mec = mec
    while _mec[MITO_CONFIG_VERSION] in mec_upgrade_functions:
        _mec = mec_upgrade_functions[_mec[MITO_CONFIG_VERSION]](_mec)

    return _mec

def is_env_variable_set_to_true(env_variable: Union[str, int, bool]) -> bool: 
    truth_values = ['True', 'true', True, 1, '1', 't', 'T']
    return env_variable in truth_values

def create_mec_from_environment_variables() -> Optional[Dict[str, Any]]:
    """
    Creates a Mito Enterprise Config object from the environment variables
    """
    config_version = os.environ.get(MITO_CONFIG_VERSION)

    if config_version is None:
        return None

    mec: Dict[str, Any] = {}
    for key in MEC_VERSION_KEYS[config_version]:
        mec[key] = os.environ.get(key)

    return mec

def get_enterprise_from_config(mito_config_enterprise: Optional[str], mito_config_enterprise_temp_license: Optional[str]) -> bool:
    
    if mito_config_enterprise is None and mito_config_enterprise_temp_license is None:
        return False
    
    enterprise = mito_config_enterprise is not None and is_env_variable_set_to_true(mito_config_enterprise)
    if enterprise:
        return enterprise
    
    if mito_config_enterprise_temp_license is None:
        return False
    
    enterprise_temp_license_string = mito_config_enterprise_temp_license
    if enterprise_temp_license_string is not None:
        enterprise_temp_license_date = decode_license_to_date(enterprise_temp_license_string)
        enterprise_temp_license = enterprise_temp_license_date > datetime.date.today()
        return enterprise_temp_license
    
    return False


class MitoConfig:
    """
    The MitoConfig class is repsonsible for reading the settings from the 
    environment variables and returning them as the most updated version of the 
    mito_enterprise_configuration object that is used by the rest of the app. 

    If the environment variables does not exist or does not set every configuration option, 
    the MitoConfig class sets defaults. 
    """
    def __init__(self):
        mec_potentially_outdated = create_mec_from_environment_variables()
        self.mec = upgrade_mito_enterprise_configuration(mec_potentially_outdated)

        if self.mec is not None:
            log('loaded_mito_enterprise_config')

    @property
    def version(self) -> str:
        if self.mec is None or self.mec[MITO_CONFIG_VERSION] is None:
            return '2' # NOTE: update this to be the most recent version, when we bump the version
        return self.mec[MITO_CONFIG_VERSION]
    
    @property
    def support_email(self) -> str:
        if self.mec is None or self.mec[MITO_CONFIG_SUPPORT_EMAIL] is None:
            return DEFAULT_MITO_CONFIG_SUPPORT_EMAIL
        return self.mec[MITO_CONFIG_SUPPORT_EMAIL]

    @property
    def disable_tours(self) -> bool:
        if self.mec is None or self.mec[MITO_CONFIG_DISABLE_TOURS] is None:
            return False

        disable_tours = is_env_variable_set_to_true(self.mec[MITO_CONFIG_DISABLE_TOURS])
        return disable_tours

    @property
    def enable_snowflake_import(self) -> bool:
        if self.mec is None or self.mec[MITO_CONFIG_FEATURE_ENABLE_SNOWFLAKE_IMPORT] is None:
            return False

        enable_snowflake_import = is_env_variable_set_to_true(self.mec[MITO_CONFIG_FEATURE_ENABLE_SNOWFLAKE_IMPORT])
        return enable_snowflake_import
    
    @property
    def display_ai_transform(self) -> bool:
        """
        We display AI transformation on the frontend if:
        1. The user is not on Mito Enterprise
        2. MITO_CONFIG_FEATURE_DISPLAY_AI_TRANSFORMATION is not False

        Note that this means that MITO_CONFIG_FEATURE_DISPLAY_AI_TRANSFORMATION not being set
        means we get a default value is True.
        """
        
        if self.mec is not None:
            raw_display_ai_transform = self.mec[MITO_CONFIG_FEATURE_DISPLAY_AI_TRANSFORMATION]
            display_ai_transform = is_env_variable_set_to_true(raw_display_ai_transform)

            if raw_display_ai_transform is not None:
                return display_ai_transform
            
        on_enterprise = is_enterprise()

        if on_enterprise:
            return False
        
        return True
    
    @property
    def display_snowflake_import(self) -> bool:
        """
        We display the snowflake import on the frontend if MITO_CONFIG_FEATURE_DISPLAY_SNOWFLAKE_IMPORT is not set to False.

        Note: That this just determines if the feature is __visible__ in the frontend, not if its enabled/disabled.
        To set the snowflake import as enabled, use MITO_CONFIG_FEATURE_ENABLE_SNOWFLAKE_IMPORT or be on Enterprise.
        """
        
        if self.mec is not None:
            raw_display_snowflake_import = self.mec[MITO_CONFIG_FEATURE_DISPLAY_SNOWFLAKE_IMPORT]
            display_snowflake_import = is_env_variable_set_to_true(raw_display_snowflake_import)
            return display_snowflake_import if raw_display_snowflake_import is not None else True # default to True
        
        return True
        
    @property
    def display_scheduling(self) -> bool:
        """
        We display the scheduling functionality on the frontend if MITO_CONFIG_FEATURE_DISPLAY_SCHEDULING is not set to False.

        Note: That this just determines if the feature is __visible__ in the frontend, not if its enabled/disabled.
        """
        
        if self.mec is not None:
            raw_display_scheduling = self.mec[MITO_CONFIG_FEATURE_DISPLAY_SCHEDULING]
            display_scheduling = is_env_variable_set_to_true(raw_display_scheduling)
            return display_scheduling if raw_display_scheduling is not None else True # default to True
        
        return True
    
    @property
    def display_code_options(self) -> bool:
        """
        We display the code snippets functionality on the frontend if MITO_CONFIG_FEATURE_CODE_OPTIONS is not set to False.

        Note: That this just determines if the feature is __visible__ in the frontend, not if its enabled/disabled.
        """
        
        if self.mec is not None:
            raw_display_code_options = self.mec[MITO_CONFIG_FEATURE_DISPLAY_CODE_OPTIONS]
            display_scheduling = is_env_variable_set_to_true(raw_display_code_options)
            return display_scheduling if raw_display_code_options is not None else True # default to True
        
        return True

    def _get_code_snippets_version(self) -> Optional[str]:
        if self.mec is None or self.mec[MITO_CONFIG_CODE_SNIPPETS_VERSION] is None:
            return None
        return self.mec[MITO_CONFIG_CODE_SNIPPETS_VERSION]

    def _get_code_snippets_url(self) -> Optional[str]:
        if self.mec is None or self.mec[MITO_CONFIG_CODE_SNIPPETS_URL] is None:
            return None
        return self.mec[MITO_CONFIG_CODE_SNIPPETS_URL]

    def _get_code_snippets_support_email(self) -> Optional[str]:
        if self.mec is None or self.mec[MITO_CONFIG_CODE_SNIPPETS_SUPPORT_EMAIL] is None:
            return None
        return self.mec[MITO_CONFIG_CODE_SNIPPETS_SUPPORT_EMAIL]

    @property
    def code_snippets(self) -> Optional[CodeSnippetEnvVars]:
        code_snippets_version = self._get_code_snippets_version()
        code_snippets_url = self._get_code_snippets_url()
        code_snippets_support_email = self._get_code_snippets_support_email()

        if code_snippets_version is None and code_snippets_url is None:
            return None

        if code_snippets_version != '1':
            log('mito_config_error', {'mito_config_error_reason': 'mito_config_code_snippet_version not set'})
            raise ValueError(
                "The code snippet environment variables are configured improperly. The MITO_CONFIG_CODE_SNIPPETS_URL environment variable is set, but the MITO_CONFIG_CODE_SNIPPETS_VERSION environment variable is not set to '1'.)"
            )

        if code_snippets_url is None:
            log('mito_config_error', {'mito_config_error_reason': 'mito_config_code_snippets_url not set'})
            raise ValueError(
                "The code snippet environment variables are configured improperly. The MITO_CONFIG_CODE_SNIPPETS_VERSION environment variable is set, but the MITO_CONFIG_CODE_SNIPPETS_URL environment variable is not set."
            )  

        code_snippets: CodeSnippetEnvVars = {
            'MITO_CONFIG_CODE_SNIPPETS_VERSION': code_snippets_version,
            'MITO_CONFIG_CODE_SNIPPETS_URL': code_snippets_url, 
            'MITO_CONFIG_CODE_SNIPPETS_SUPPORT_EMAIL': code_snippets_support_email 
        }
        return code_snippets
    
    @property
    def llm_url(self) -> Optional[str]:
        if self.mec is None or self.mec[MITO_CONFIG_LLM_URL] is None:
            return None
        return self.mec[MITO_CONFIG_LLM_URL]
    
    @property
    def analytics_url(self) -> Optional[str]:
        if self.mec is None or self.mec[MITO_CONFIG_ANALYTICS_URL] is None:
            return None
        return self.mec[MITO_CONFIG_ANALYTICS_URL]
    
    @property
    def log_server_url(self) -> Optional[str]:
        if self.mec is None or self.mec[MITO_CONFIG_LOG_SERVER_URL] is None:
            return None
        return self.mec[MITO_CONFIG_LOG_SERVER_URL]
    
    @property
    def log_server_batch_interval(self) -> Optional[int]:
        if self.mec is None or self.mec[MITO_CONFIG_LOG_SERVER_BATCH_INTERVAL] is None:
            return None
        return int(self.mec[MITO_CONFIG_LOG_SERVER_BATCH_INTERVAL])

    @property
    def custom_sheet_functions_path(self) -> Optional[str]:
        if self.mec is None or self.mec[MITO_CONFIG_CUSTOM_SHEET_FUNCTIONS_PATH] is None:
            return None
        else:
            return self.mec[MITO_CONFIG_CUSTOM_SHEET_FUNCTIONS_PATH]

    @property
    def custom_importers_path(self) -> Optional[str]:
        if self.mec is None or self.mec[MITO_CONFIG_CUSTOM_IMPORTERS_PATH] is None:
            return None
        else:
            return self.mec[MITO_CONFIG_CUSTOM_IMPORTERS_PATH]
    
    @property
    def feature_telemetry(self) -> bool:
        if self.mec is None or self.mec[MITO_CONFIG_FEATURE_TELEMETRY] is None:
            return True

        feature_telemetry = is_env_variable_set_to_true(self.mec[MITO_CONFIG_FEATURE_TELEMETRY])
        return feature_telemetry
    
    @property
    def pro(self) -> bool:
        if self.mec is None or self.mec[MITO_CONFIG_PRO] is None:
            return False

        pro = is_env_variable_set_to_true(self.mec[MITO_CONFIG_PRO])
        return pro
    
    @property
    def enterprise(self) -> bool:
        if self.mec is None or (
            self.mec[MITO_CONFIG_ENTERPRISE] is None
            and self.mec[MITO_CONFIG_ENTERPRISE_TEMP_LICENSE] is None
        ):
            return False
        
        return get_enterprise_from_config(
            self.mec[MITO_CONFIG_ENTERPRISE],
            self.mec[MITO_CONFIG_ENTERPRISE_TEMP_LICENSE]
        )

    # Add new mito configuration options here ...

    @property
    def mito_config_dict(self) -> Dict[str, Any]:
        return {
            MITO_CONFIG_VERSION: self.version,
            MITO_CONFIG_SUPPORT_EMAIL: self.support_email,
            MITO_CONFIG_DISABLE_TOURS: self.disable_tours,
            MITO_CONFIG_CODE_SNIPPETS: self.code_snippets,
            MITO_CONFIG_FEATURE_ENABLE_SNOWFLAKE_IMPORT: self.enable_snowflake_import,
            MITO_CONFIG_FEATURE_DISPLAY_SNOWFLAKE_IMPORT: self.display_snowflake_import,
            MITO_CONFIG_FEATURE_DISPLAY_AI_TRANSFORMATION: self.display_ai_transform,
            MITO_CONFIG_FEATURE_DISPLAY_SCHEDULING: self.display_scheduling,
            MITO_CONFIG_FEATURE_DISPLAY_CODE_OPTIONS: self.display_code_options,
            MITO_CONFIG_LLM_URL: self.llm_url,
            MITO_CONFIG_ANALYTICS_URL: self.analytics_url,
            MITO_CONFIG_LOG_SERVER_URL: self.log_server_url,
            MITO_CONFIG_LOG_SERVER_BATCH_INTERVAL: self.log_server_batch_interval,
            MITO_CONFIG_FEATURE_TELEMETRY: self.feature_telemetry,
            MITO_CONFIG_CUSTOM_SHEET_FUNCTIONS_PATH: self.custom_sheet_functions_path,
            MITO_CONFIG_CUSTOM_IMPORTERS_PATH: self.custom_importers_path,
            MITO_CONFIG_PRO: self.pro,
            MITO_CONFIG_ENTERPRISE: self.enterprise
        }




#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Contains tests to make sure that the mito analytics test is
performing correctly
"""

import json
import os
import sys
import time
from unittest.mock import MagicMock, patch
import requests

import pandas as pd

from mitosheet.enterprise.mito_config import (
    MITO_CONFIG_LOG_SERVER_BATCH_INTERVAL, MITO_CONFIG_LOG_SERVER_URL,
    MITO_CONFIG_VERSION)
from mitosheet.telemetry.telemetry_utils import PRINT_LOGS
from mitosheet.tests.test_mito_config import \
    delete_all_mito_config_environment_variables
from mitosheet.tests.test_utils import (create_mito_wrapper,
                                        create_mito_wrapper_with_data)

try:
    from pandas import __version__ as pandas_version
except:
    pandas_version = 'no pandas'
try:
    # Format version_python as "3.8.5"
    version_python = '.'.join([str(x) for x in sys.version_info[:3]])
except:
    version_python = 'no python'

URL = "https://url?" 

def test_not_printing_logs():
    assert PRINT_LOGS is False

def test_log_uploader_single_edit_event():
    
    os.environ[MITO_CONFIG_VERSION] = "2"
    os.environ[MITO_CONFIG_LOG_SERVER_URL] =  f"{URL}"
    os.environ[MITO_CONFIG_LOG_SERVER_BATCH_INTERVAL] = "0"
    
    mito = create_mito_wrapper_with_data([123])

    with patch('requests.post') as mock_post:
        mito.add_column(0, 'B')

        print(mock_post.call_args_list)
        log_call = mock_post.call_args_list[0]

        # Get the URL from the log call
        actual_url = log_call[0][0]
        assert actual_url == URL

        # From the call object, get the payload 
        # that was passed to requests.post
        data = log_call[1]['data']
        log_payload = json.loads(data)[0]

        assert len(log_payload) == 9
        assert log_payload["params_sheet_index"] == 0
        assert log_payload["params_column_header"] is not None
        assert log_payload["params_column_header_index"] == -1
        assert log_payload["params_public_interface_version"] == 3
        assert log_payload["version_python"] == version_python
        assert log_payload["version_pandas"] == pandas_version
        assert log_payload["version_mito"] is not None
        assert log_payload["timestamp_gmt"] is not None
        assert log_payload["event"] == "add_column_edit"

    delete_all_mito_config_environment_variables()

def test_log_uploader_multiple_edit_events():
    
    os.environ[MITO_CONFIG_VERSION] = "2"
    os.environ[MITO_CONFIG_LOG_SERVER_URL] =  f"{URL}"
    os.environ[MITO_CONFIG_LOG_SERVER_BATCH_INTERVAL] = "1"
    
    mito = create_mito_wrapper_with_data([123])

    mito.add_column(0, 'B')
    mito.delete_columns(0, ['B'])

    # wait 1 second, so we don't get these uploads
    time.sleep(1)

    with patch('requests.post') as mock_post:

        mito.add_column(0, 'C')
        time.sleep(2)

        # Even though multiple events occured, there should be only one upload
        assert len(mock_post.call_args_list) == 1

        log_call = mock_post.call_args_list[0]

        # Get the URL from the log call
        actual_url = log_call[0][0]
        assert actual_url == URL

        # From the call object, get the payload 
        # that was passed to requests.post
        data = log_call[1]['data']
        add_column_log_event = json.loads(data)[0]

        assert len(add_column_log_event) == 9
        assert add_column_log_event["params_sheet_index"] == 0
        assert add_column_log_event["params_column_header"] is not None
        assert add_column_log_event["params_column_header_index"] == -1
        assert add_column_log_event["params_public_interface_version"] == 3
        assert add_column_log_event["version_python"] == version_python
        assert add_column_log_event["version_pandas"] == pandas_version
        assert add_column_log_event["version_mito"] is not None
        assert add_column_log_event["timestamp_gmt"] is not None
        assert add_column_log_event["event"] == "add_column_edit"

        delete_columns_log_event = json.loads(data)[1]
        assert len(delete_columns_log_event) == 8
        assert delete_columns_log_event["params_sheet_index"] == 0
        assert len(delete_columns_log_event["params_column_ids"]) == 1
        assert delete_columns_log_event["params_public_interface_version"] == 3
        assert delete_columns_log_event["version_python"] == version_python
        assert delete_columns_log_event["version_pandas"] == pandas_version
        assert delete_columns_log_event["version_mito"] is not None
        assert delete_columns_log_event["timestamp_gmt"] is not None
        assert delete_columns_log_event["event"] == "delete_column_edit"

    delete_all_mito_config_environment_variables()

def test_log_uploader_error_events():
    
    os.environ[MITO_CONFIG_VERSION] = "2"
    os.environ[MITO_CONFIG_LOG_SERVER_URL] =  f"{URL}"
    os.environ[MITO_CONFIG_LOG_SERVER_BATCH_INTERVAL] = "0"
    
    mito = create_mito_wrapper(pd.DataFrame({'A': [1, 2, 3]}))

    with patch('requests.post') as mock_post:

        mito.set_formula('=INVALID_FUNCTION()', 0, 'A')
        time.sleep(.5)

        assert len(mock_post.call_args_list) == 1

        log_call = mock_post.call_args_list[0]

        # Get the URL from the log call
        actual_url = log_call[0][0]
        assert actual_url == URL

        # From the call object, get the payload 
        # that was passed to requests.post
        data = log_call[1]['data']
        log_event = json.loads(data)[0]

        assert len(log_event) == 15
        assert log_event["error_traceback"] is not None
        assert log_event["error_traceback_last_line"] is not None
        assert log_event["params_sheet_index"] == 0
        assert log_event["params_column_id"] is not None
        assert log_event["params_formula_label"] is not None
        assert log_event["params_index_labels_formula_is_applied_to"] is not None
        assert log_event["params_public_interface_version"] == 3
        assert log_event["params_new_formula"] == '=INVALID_FUNCTION()'
        assert log_event["version_python"] == version_python
        assert log_event["version_pandas"] == pandas_version
        assert log_event["version_mito"] is not None
        assert log_event["timestamp_gmt"] is not None
        assert log_event["event"] == "error"
        assert log_event["params_failed_log_event"] == "set_column_formula_edit_failed"
        assert log_event["error_severity_code"] == "10"


    delete_all_mito_config_environment_variables()
    

def test_log_uploader_mitosheet_rendered():
    
    os.environ[MITO_CONFIG_VERSION] = "2"
    os.environ[MITO_CONFIG_LOG_SERVER_URL] = f"{URL}"
    os.environ[MITO_CONFIG_LOG_SERVER_BATCH_INTERVAL] = "0"
    
    mito = create_mito_wrapper(pd.DataFrame({'A': [1, 2, 3]}))

    with patch('requests.post') as mock_post:
        mito.mito_backend.receive_message({
            'params': {'user_agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/17.2.1 Safari/605.1.15'}, 
            'event': 'log_event', 
            'type': 'mitosheet_rendered', 
            'id': '_pt4wgfw0z'
        })
        time.sleep(.5)

        assert len(mock_post.call_args_list) == 1

        log_call = mock_post.call_args_list[0]

        # Get the URL from the log call
        actual_url = log_call[0][0]
        assert actual_url == URL

        # From the call object, get the payload 
        # that was passed to requests.post
        data = log_call[1]['data']
        log_event = json.loads(data)[0]

        assert len(log_event) == 6
        assert log_event["params_user_agent"] == "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/17.2.1 Safari/605.1.15"
        assert log_event["version_python"] == version_python
        assert log_event["version_pandas"] == pandas_version
        assert log_event["version_mito"] is not None
        assert log_event["timestamp_gmt"] is not None
        assert log_event["event"] == "mitosheet_rendered"

    delete_all_mito_config_environment_variables()

def test_log_uploaded_frontend_render_failed():
    os.environ[MITO_CONFIG_VERSION] = "2"
    os.environ[MITO_CONFIG_LOG_SERVER_URL] = f"{URL}"
    os.environ[MITO_CONFIG_LOG_SERVER_BATCH_INTERVAL] = "0"
    
    mito = create_mito_wrapper(pd.DataFrame({'A': [1, 2, 3]}))

    with patch('requests.post') as mock_post:
        mito.mito_backend.receive_message({
            'params': {'error_stack': 'TEST ERROR STATCK', 'error_message': 'TEST ERROR MESSAGE'}, 
            'event': 'log_event', 
            'type': 'frontend_render_failed', 
            'id': '_pt4wgfw0z'
        })

        time.sleep(1)

        assert len(mock_post.call_args_list) == 1

        log_call = mock_post.call_args_list[0]

        # Get the URL from the log call
        actual_url = log_call[0][0]
        assert actual_url == URL

        # From the call object, get the payload 
        # that was passed to requests.post
        data = log_call[1]['data']
        log_event = json.loads(data)[0]

        assert log_event['params_error_stack'] is not None
        assert log_event['params_error_message'] is not None
        assert log_event['event'] == 'frontend_render_failed'
        assert log_event['error_severity_code'] == '50'

    delete_all_mito_config_environment_variables()


def test_log_uploader_long_interval_does_not_trigger_upload():
    
    os.environ[MITO_CONFIG_VERSION] = "2"
    os.environ[MITO_CONFIG_LOG_SERVER_URL] =  f"{URL}"
    os.environ[MITO_CONFIG_LOG_SERVER_BATCH_INTERVAL] = "100"
    
    mito = create_mito_wrapper(pd.DataFrame({'A': [1, 2, 3]}))

    with patch('requests.post') as mock_post:

        mito.add_column(0, 'B')
        mito.add_column(0, 'C')
        mito.add_column(0, 'D')

        assert len(mock_post.call_args_list) == 0

    delete_all_mito_config_environment_variables()

def test_log_uploader_logs_failure_prints():

    os.environ[MITO_CONFIG_VERSION] = "2"
    os.environ[MITO_CONFIG_LOG_SERVER_URL] =  f"{URL}"
    os.environ[MITO_CONFIG_LOG_SERVER_BATCH_INTERVAL] = "1"

    mito = create_mito_wrapper(pd.DataFrame({'A': [1, 2, 3]}))

    with patch('requests.post') as mock_post:
        mock_post.side_effect = [requests.exceptions.RequestException, None]

        # Capture the logging debug output
        with patch('builtins.print') as mock_print:
            mito.add_column(0, 'B')
            time.sleep(1.5)

            assert len(mock_print.call_args_list) == 1
            assert 'Log upload failed with error:' in mock_print.call_args_list[0][0][0]

    delete_all_mito_config_environment_variables()

def test_log_uploader_tries_failed_logs_again():

    os.environ[MITO_CONFIG_VERSION] = "2"
    os.environ[MITO_CONFIG_LOG_SERVER_URL] =  f"{URL}"
    os.environ[MITO_CONFIG_LOG_SERVER_BATCH_INTERVAL] = "1"

    mito = create_mito_wrapper(pd.DataFrame({'A': [1, 2, 3]}))

    with patch('requests.post') as mock_post:
        mock_post.side_effect = [requests.exceptions.RequestException, None]
        mito.add_column(0, 'B')
        time.sleep(1.5)

    with patch('requests.post') as mock_post:
        mito.add_column(0, 'C')
        time.sleep(1.5)
        assert len(mock_post.call_args_list) == 1
        logs = json.loads(mock_post.call_args_list[0][1]['data'])
        assert len(logs) == 2
        # Check that the payload is different, indicating that the first
        # payload was not uploaded the first time
        assert logs[0]['params_column_header'] != logs[1]['params_column_header']

    delete_all_mito_config_environment_variables()

def test_mito_log_uploader_exponential_backoff():
    
    os.environ[MITO_CONFIG_VERSION] = "2"
    os.environ[MITO_CONFIG_LOG_SERVER_URL] =  f"{URL}"
    os.environ[MITO_CONFIG_LOG_SERVER_BATCH_INTERVAL] = "1"

    mito = create_mito_wrapper(pd.DataFrame({'A': [1, 2, 3]}))

    with patch('requests.post') as mock_post:
        mock_post.side_effect = [requests.exceptions.RequestException, None]
        mito.add_column(0, 'B')
        time.sleep(1.5)
        assert mito.mito_backend.steps_manager.mito_log_uploader is not None and mito.mito_backend.steps_manager.mito_log_uploader.current_log_interval == 2

    with patch('requests.post') as mock_post:
        mito.add_column(0, 'C')
        time.sleep(2)
        assert mito.mito_backend.steps_manager.mito_log_uploader is not None and mito.mito_backend.steps_manager.mito_log_uploader.current_log_interval == 1

    delete_all_mito_config_environment_variables()


def test_mito_log_uploader_prints_on_server_error():

    os.environ[MITO_CONFIG_VERSION] = "2"
    os.environ[MITO_CONFIG_LOG_SERVER_URL] =  f"{URL}"
    os.environ[MITO_CONFIG_LOG_SERVER_BATCH_INTERVAL] = "1"

    mito = create_mito_wrapper(pd.DataFrame({'A': [1, 2, 3]}))

    mock_response = MagicMock(spec=requests.Response)
    mock_response.status_code = 400
    mock_response.raise_for_status.side_effect = requests.HTTPError

    with patch('requests.post', return_value=mock_response) as mock_post:
        mito.add_column(0, 'C')
        time.sleep(1.5)

    assert mito.mito_backend.steps_manager.mito_log_uploader is not None and mito.mito_backend.steps_manager.mito_log_uploader.current_log_interval == 2

    delete_all_mito_config_environment_variables()



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.


from typing import TYPE_CHECKING, Any, List, Optional, Tuple

from mitosheet.types import ParamSubtype, ParamType, ParamValue

if TYPE_CHECKING:
    from mitosheet.state import State
else:
    State = Any

class CodeChunk:
    """
    A CodeChunk is the a abstract base class that can be inhereited from
    to create a code generating object. For example, an AddColumnCodeChunk
    with the correct parameters and states will generate code that adds
    a column with a specific column header to a specific dataframe.

    When a step is transpiled, it generates a list of code chunks. In turn,
    we can optimize these code chunks to allow us to write the most minimal
    amount of code. 
    """

    def __init__(self, 
        prev_state: State,
    ):
        self.prev_state = prev_state

        self.optional_code_that_successfully_executed: Tuple[List[str], List[str]] = ([], [])

    def __repr__(self) -> str:
        members = [
            (attr, getattr(self, attr)) for attr in dir(self) if not callable(getattr(self, attr)) and not attr.startswith("__")
            and ('sheet_index' in attr or 'file_name' in attr)
        ]
        return f"{self.__class__.__name__}({members})"

    def get_display_name(self) -> str:
        """Returns a short name to display for this CodeChunk"""
        raise NotImplementedError('Implement in subclass')
    
    def get_description_comment(self) -> str:
        """Returns a detailed comment explaing what happened in this CodeChunk"""
        raise NotImplementedError('Implement in subclass')

    def get_code(self) -> Tuple[List[str], List[str]]:
        """
        Returns a Tuple of lists of code strings that this code chunk executes. 
        The first list is the code used to perfrom the operation. The second list 
        is the imports required for the code. ie: import pandas as pd
        """
        raise NotImplementedError('Implement in subclass')
    
    def get_optional_code_that_successfully_executed(self) -> Tuple[List[str], List[str]]:
        """
        Consider a code chunk like PivotCodeChunk. The main work it does is creating a new
        dataframe that has a pivot table inside of it. 

        However, if the user is editing a pivot table, we may want to replay some edits on
        top of it. Our strategy for this is to go and find all of the code chunks that edit
        the OG pivot table, and then to try to execute the code from each of these code 
        chunks on top of the new edited pivot table. 

        For example if you:
        1. Make a pivot
        2. Add a column to the pivot
        3. Rename a column in the pivot
        4. Filter the renamed column
        5. Edit the pivot

        Then the new column, rename, and filter will be replayed on top of the edited pivot. 
        However, if a code chunk fails (e.g. the rename fails), then we don't execute later 
        code lines, as to avoid getting the pivot into a super strange state. 

        Thus, we have a new type of code for the code chunk. The `get_code` results _must_
        all execute correctly, for the step to terminate correctly. If any of the lines of
        code from get_code fail, then the entire step is rolled back and not created. 

        However, optional code lines can fail and the step still be correctly executed. Thus, 
        we have to have a notion of a) optional code lines, and b) the code chunk needs to know
        which of these optional lines was executed correctly. 

        In the above example, if the rename step failed, then this function would return the 
        code to add a column to the pivot table, and not the code to rename or filter.
        """
        return self.optional_code_that_successfully_executed
        
    # TODO: we could add a function that returns a list of params and execution
    # data that you're allowed to reference, and then check this when you create
    # the step, for strong typing!

    def params_match(self, other_code_chunk: "CodeChunk", param_keys: List[str]) -> bool:
        """
        Given a different code chunk, and a list of keys to check, returns True if
        all the given keys match in the params. A useful utility for checking if 
        CodeChunks are compatible for combination.
        """
        for key in param_keys:
            if self.__dict__[key] != other_code_chunk.__dict__[key]:
                return False
        return True

    def get_created_sheet_indexes(self) -> Optional[List[int]]:
        """
        Dataframe deletes allow us to optimize a lot of code, we allow steps to
        optionally say that they only create some specific list of sheet_indexes.

        If this function returns a sheet index, and later dataframe delete steps
        delete this sheet index, then we will optimize out this step as well as
        the deleting of the dataframe.

        NOTE: if this funciton returns None, it is this CodeChunk saying that
        it cannot do any optimization with dataframe delete - which we do by
        default.
        """
        return None
    
    def get_edited_sheet_indexes(self) -> Optional[List[int]]:
        """
        Dataframe deletes allow us to optimize a lot of code, we allow steps to
        optionally say that they only edit some specific list of sheet_indexes.

        This allows us to easily optimize out these steps if the dataframe they 
        are editing is then deleted.

        NOTE: if this funciton returns None, it is this CodeChunk saying that
        it cannot do any optimization with dataframe delete - which we do by
        default.
        """
        return None
    
    def get_source_sheet_indexes(self) -> Optional[List[int]]:
        """
        This funciton returns a list of sheet indexes that go into
        creating any created_sheet_indexes. 

        If there are no created_sheet_indexes, than this function can
        be ignored. If there are created_sheet_indexes, than an empty
        list means there were no source dataframes -- e.g. from a CSV 
        import from a file.
        """
        return []

    
    def combine_right(self, other_code_chunk: "CodeChunk") -> Optional["CodeChunk"]:
        """
        Given a list of CodeChunks [A, B], combine right called on A with
        B as a parameter will check if A and B can be combined into a new
        CodeChunk. 

        If they cannot be combined, None will be returned. If they can be
        combined, the new combined CodeChunk will be returned, and thus
        [A, B] goes to [A.combine_right(B)]
        """
        return None
    
    def combine_left(self, other_code_chunk: "CodeChunk") -> Optional["CodeChunk"]:
        """Given a list of CodeChunks [A, B], combine right called on B with
        A as a parameter will check if A and B can be combined into a new
        CodeChunk. 

        If they cannot be combined, None will be returned. If they can be
        combined, the new combined CodeChunk will be returned, and thus
        [A, B] goes to [B.combine_right(A)]

        NOTE: combine_lefts are only done after the combine_rights. You might want
        to do one vs. the other depending on how natural it is to express. For example,
        because deleting a dataframe delete can remove a ton of other steps, expressing
        it as a combine_left is much more natural as it results in this optimization
        code all living in one location.
        """
        return None
    
    def get_parameterizable_params(self) -> List[Tuple[ParamValue, ParamType, ParamSubtype]]:
        """
        Returns a list of parameters that can be parameterized. This is used
        for the parameterization UI.

        The tuple is (param_name, param_type, param_subtype).
        """
        return []
    
    def can_be_reordered_with(self, code_chunk: "CodeChunk") -> bool:
        """
        Returns true if the passed code chunk can be moved from
        before to after (or after to before) with this code chunk. 

        Note that this function is meant to be conservative. It should
        only return True if the reordering of the code chunks _positively_
        does not break generated code. 


        """

        created_sheet_indexes = self.get_created_sheet_indexes()
        edited_sheet_indexes = self.get_edited_sheet_indexes()
        source_sheet_indexes = self.get_source_sheet_indexes()
        other_edited_indexes = code_chunk.get_edited_sheet_indexes()

        # Don't reorder if the code chunk is editing what this code chunk created
        if created_sheet_indexes is not None and other_edited_indexes is not None and any(index in created_sheet_indexes for index in other_edited_indexes):
            return False

        # Don't reorder if the code chunks are editing the same sheet
        if edited_sheet_indexes is not None and other_edited_indexes is not None and set(edited_sheet_indexes) == set(other_edited_indexes):
            return False
        
        # Don't reorder if the other code chunk edits where this code chunk pulls from
        if source_sheet_indexes is not None and other_edited_indexes is not None and any(index in source_sheet_indexes for index in other_edited_indexes):
            return False

        return True